# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import List, Optional, Tuple
from click import command, option, argument, Choice

from lean.click import LeanCommand, PathParameter
from lean.constants import DEFAULT_ENGINE_IMAGE, LEAN_ROOT_PATH
from lean.container import container, Logger
from lean.models.utils import DebuggingMethod
from lean.models.cli import cli_data_downloaders, cli_addon_modules
from lean.components.util.json_modules_handler import build_and_configure_modules, non_interactive_config_build_for_name
from lean.models.click_options import options_from_json, get_configs_for_options

# The _migrate_* methods automatically update launch configurations for a given debugging method.
#
# Occasionally we make changes which require updated launch configurations.
# Projects which are created after these update have the correct configuration already,
# but projects created before that need changes.
#
# These methods checks if the project has outdated configurations, and if so, update them to keep it working.


def _migrate_python_pycharm(logger: Logger, project_dir: Path) -> None:
    from os import path
    from click import Abort

    workspace_xml_path = project_dir / ".idea" / "workspace.xml"
    if not workspace_xml_path.is_file():
        return

    xml_manager = container.xml_manager
    current_content = xml_manager.parse(workspace_xml_path.read_text(encoding="utf-8"))

    config = current_content.find('.//configuration[@name="Debug with Lean CLI"]')
    if config is None:
        return

    path_mappings = config.find('.//PathMappingSettings/option[@name="pathMappings"]/list')
    if path_mappings is None:
        return

    made_changes = False
    has_library_mapping = False

    library_dir = container.lean_config_manager.get_cli_root_directory() / "Library"

    if library_dir.is_dir():
        library_dir = f"$PROJECT_DIR$/{path.relpath(library_dir, project_dir)}".replace("\\", "/")
    else:
        library_dir = None

    for mapping in path_mappings.findall(".//mapping"):
        if mapping.get("local-root") == "$PROJECT_DIR$" and mapping.get("remote-root") == LEAN_ROOT_PATH:
            mapping.set("remote-root", "/LeanCLI")
            made_changes = True

        if library_dir is not None \
            and mapping.get("local-root") == library_dir \
            and mapping.get("remote-root") == "/Library":
            has_library_mapping = True

    if library_dir is not None and not has_library_mapping:
        library_mapping = xml_manager.parse("<mapping/>")
        library_mapping.set("local-root", library_dir)
        library_mapping.set("remote-root", "/Library")
        path_mappings.append(library_mapping)
        made_changes = True

    if made_changes:
        workspace_xml_path.write_text(xml_manager.to_string(current_content), encoding="utf-8")

        logger = container.logger
        logger.warn("Your run configuration has been updated to work with the latest version of LEAN")
        logger.warn("Please restart the debugger in PyCharm and run this command again")

        raise Abort()


def _migrate_python_vscode(project_dir: Path) -> None:
    from json import dumps, loads
    launch_json_path = project_dir / ".vscode" / "launch.json"
    if not launch_json_path.is_file():
        return

    current_content = loads(launch_json_path.read_text(encoding="utf-8"))
    if "configurations" not in current_content or not isinstance(current_content["configurations"], list):
        return

    config = next((c for c in current_content["configurations"] if c["name"] == "Debug with Lean CLI"), None)
    if config is None:
        return

    made_changes = False
    has_library_mapping = False

    library_dir = container.lean_config_manager.get_cli_root_directory() / "Library"
    if not library_dir.is_dir():
        library_dir = None

    for mapping in config["pathMappings"]:
        if mapping["localRoot"] == "${workspaceFolder}" and mapping["remoteRoot"] == LEAN_ROOT_PATH:
            mapping["remoteRoot"] = "/LeanCLI"
            made_changes = True

        if library_dir is not None and mapping["localRoot"] == str(library_dir) and mapping["remoteRoot"] == "/Library":
            has_library_mapping = True

    if library_dir is not None and not has_library_mapping:
        config["pathMappings"].append({
            "localRoot": str(library_dir),
            "remoteRoot": "/Library"
        })
        made_changes = True

    if made_changes:
        launch_json_path.write_text(dumps(current_content, indent=4), encoding="utf-8")


def _migrate_csharp_rider(logger: Logger, project_dir: Path) -> None:
    from click import Abort

    made_changes = False
    xml_manager = container.xml_manager

    for dir_name in [f".idea.{project_dir.stem}", f".idea.{project_dir.stem}.dir"]:
        workspace_xml_path = project_dir / ".idea" / dir_name / ".idea" / "workspace.xml"
        if not workspace_xml_path.is_file():
            continue

        current_content = xml_manager.parse(workspace_xml_path.read_text(encoding="utf-8"))

        run_manager = current_content.find(".//component[@name='RunManager']")
        if run_manager is None:
            continue

        config = run_manager.find(".//configuration[@name='Debug with Lean CLI']")
        if config is None:
            continue

        run_manager.remove(config)

        workspace_xml_path.write_text(xml_manager.to_string(current_content), encoding="utf-8")
        made_changes = True

    if container.project_manager.generate_rider_config(project_dir) or made_changes:
        logger.warn("Your debugger configuration has been updated to work with the latest version of LEAN")
        logger.warn("Please restart Rider and start debugging again")
        logger.warn(
            "See https://www.lean.io/docs/v2/lean-cli/backtesting/debugging#05-C-and-Rider for the updated instructions")

        raise Abort()


def _migrate_csharp_vscode(project_dir: Path) -> None:
    from json import dumps, loads
    launch_json_path = project_dir / ".vscode" / "launch.json"
    if not launch_json_path.is_file():
        return

    current_content = loads(launch_json_path.read_text(encoding="utf-8"))
    if "configurations" not in current_content or not isinstance(current_content["configurations"], list):
        return

    config = next((c for c in current_content["configurations"] if c["name"] == "Debug with Lean CLI"), None)
    if config is None:
        return

    if config["type"] != "mono" and config["processId"] != "${command:pickRemoteProcess}":
        return

    config.pop("address", None)
    config.pop("port", None)

    config["type"] = "coreclr"
    config["processId"] = "1"

    config["pipeTransport"] = {
        "pipeCwd": "${workspaceRoot}",
        "pipeProgram": "docker",
        "pipeArgs": ["exec", "-i", "lean_cli_vsdbg"],
        "debuggerPath": "/root/vsdbg/vsdbg",
        "quoteArgs": False
    }

    config["logging"] = {
        "moduleLoad": False
    }

    launch_json_path.write_text(dumps(current_content, indent=4), encoding="utf-8")


def _migrate_csharp_csproj(project_dir: Path) -> None:
    csproj_path = next((f for f in project_dir.rglob("*.csproj")), None)
    if csproj_path is None:
        return

    xml_manager = container.xml_manager

    current_content = xml_manager.parse(csproj_path.read_text(encoding="utf-8"))
    if current_content.find(".//PropertyGroup/DefaultItemExcludes") is not None:
        return

    property_group = current_content.find(".//PropertyGroup")
    if property_group is None:
        property_group = xml_manager.parse("<PropertyGroup/>")
        current_content.append(property_group)

    default_item_excludes = xml_manager.parse(
        "<DefaultItemExcludes>$(DefaultItemExcludes);backtests/*/code/**;live/*/code/**;optimizations/*/code/**</DefaultItemExcludes>")
    property_group.append(default_item_excludes)

    csproj_path.write_text(xml_manager.to_string(current_content), encoding="utf-8")


@command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@option("--output",
              type=PathParameter(exists=False, file_okay=False, dir_okay=True),
              help="Directory to store results in (defaults to PROJECT/backtests/TIMESTAMP)")
@option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run the backtest in a detached Docker container and return immediately")
@option("--debug",
              type=Choice(["pycharm", "ptvsd", "vsdbg", "rider", "local-platform"], case_sensitive=False),
              help="Enable a certain debugging method (see --help for more information)")
@option("--data-provider-historical",
              type=Choice([dp.get_name() for dp in cli_data_downloaders], case_sensitive=False),
              default="Local",
              help="Update the Lean configuration file to retrieve data from the given historical provider")
@options_from_json(get_configs_for_options("backtest"))
@option("--download-data",
              is_flag=True,
              default=False,
              help="Update the Lean configuration file to download data from the QuantConnect API, alias for --data-provider-historical QuantConnect")
@option("--data-purchase-limit",
              type=int,
              help="The maximum amount of QCC to spend on downloading data during the backtest when using QuantConnect as historical data provider")
@option("--release",
              is_flag=True,
              default=False,
              help="Compile C# projects in release configuration instead of debug")
@option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@option("--python-venv",
              type=str,
              help=f"The path of the python virtual environment to be used")
@option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before running the backtest")
@option("--backtest-name",
              type=str,
              help="Backtest name")
@option("--addon-module",
              type=str,
              multiple=True,
              hidden=True)
@option("--extra-config",
              type=(str, str),
              multiple=True,
              hidden=True)
@option("--extra-docker-config",
              type=str,
              default="{}",
              help="Extra docker configuration as a JSON string. "
                   "For more information https://docker-py.readthedocs.io/en/stable/containers.html")
@option("--no-update",
              is_flag=True,
              default=False,
              help="Use the local LEAN engine image instead of pulling the latest version")
def backtest(project: Path,
             output: Optional[Path],
             detach: bool,
             debug: Optional[str],
             data_provider_historical: Optional[str],
             download_data: bool,
             data_purchase_limit: Optional[int],
             release: bool,
             image: Optional[str],
             python_venv: Optional[str],
             update: bool,
             backtest_name: str,
             addon_module: Optional[List[str]],
             extra_config: Optional[Tuple[str, str]],
             extra_docker_config: Optional[str],
             no_update: bool,
             **kwargs) -> None:
    """Backtest a project locally using Docker.

    \b
    If PROJECT is a directory, the algorithm in the main.py or Main.cs file inside it will be executed.
    If PROJECT is a file, the algorithm in the specified file will be executed.

    \b
    Go to the following url to learn how to debug backtests locally using the Lean CLI:
    https://www.lean.io/docs/v2/lean-cli/backtesting/debugging

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    from datetime import datetime
    from json import loads

    logger = container.logger
    project_manager = container.project_manager
    algorithm_file = project_manager.find_algorithm_file(Path(project))
    lean_config_manager = container.lean_config_manager
    if output is None:
        output = algorithm_file.parent / "backtests" / datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    environment_name = "backtesting"
    debugging_method = None
    if debug == "pycharm":
        debugging_method = DebuggingMethod.PyCharm
        _migrate_python_pycharm(logger, algorithm_file.parent)
    elif debug == "ptvsd":
        debugging_method = DebuggingMethod.PTVSD
        _migrate_python_vscode(algorithm_file.parent)
    elif debug == "vsdbg":
        debugging_method = DebuggingMethod.VSDBG
        _migrate_csharp_vscode(algorithm_file.parent)
    elif debug == "rider":
        debugging_method = DebuggingMethod.Rider
        _migrate_csharp_rider(logger, algorithm_file.parent)
    elif debug == "local-platform":
        debugging_method = DebuggingMethod.LocalPlatform

    if detach and debugging_method != None and debugging_method != DebuggingMethod.LocalPlatform:
        raise RuntimeError("Running a debugging session in a detached container is not supported")

    if algorithm_file.name.endswith(".cs"):
        _migrate_csharp_csproj(algorithm_file.parent)

    lean_config = lean_config_manager.get_complete_lean_config(environment_name, algorithm_file, debugging_method)

    if download_data:
        data_provider_historical = "QuantConnect"

    organization_id = container.organization_manager.try_get_working_organization_id()
    paths_to_mount = None

    if data_provider_historical is not None:
        data_provider = non_interactive_config_build_for_name(lean_config, data_provider_historical,
                                                              cli_data_downloaders, kwargs, logger, environment_name)
        data_provider.ensure_module_installed(organization_id)
        container.lean_config_manager.set_properties(data_provider.get_settings())
        paths_to_mount = data_provider.get_paths_to_mount()

    lean_config_manager.configure_data_purchase_limit(lean_config, data_purchase_limit)

    cli_config_manager = container.cli_config_manager
    project_config_manager = container.project_config_manager

    project_config = project_config_manager.get_project_config(algorithm_file.parent)
    engine_image = cli_config_manager.get_engine_image(image or project_config.get("engine-image", None))

    if str(engine_image) != DEFAULT_ENGINE_IMAGE:
        logger.warn(f'A custom engine image: "{engine_image}" is being used!')

    container.update_manager.pull_docker_image_if_necessary(engine_image, update, no_update)

    if not output.exists():
        output.mkdir(parents=True)

    # Set backtest name
    if backtest_name is not None and backtest_name != "":
        lean_config["backtest-name"] = backtest_name

    # Set extra config
    given_algorithm_id = None
    for key, value in extra_config:
        if key == "algorithm-id":
            given_algorithm_id = int(value)
        else:
            lean_config[key] = value

    output_config_manager = container.output_config_manager
    lean_config["algorithm-id"] = str(output_config_manager.get_backtest_id(output, given_algorithm_id))

    if python_venv is not None and python_venv != "":
        lean_config["python-venv"] = f'{"/" if python_venv[0] != "/" else ""}{python_venv}'

    # Configure addon modules
    build_and_configure_modules(addon_module, cli_addon_modules, organization_id, lean_config,
                                kwargs, logger, environment_name)

    lean_runner = container.lean_runner
    lean_runner.run_lean(lean_config,
                         environment_name,
                         algorithm_file,
                         output,
                         engine_image,
                         debugging_method,
                         release,
                         detach,
                         loads(extra_docker_config),
                         paths_to_mount)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Any

from lean.components.api.api_client import APIClient
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.path_manager import PathManager
from lean.components.cloud.cloud_project_manager import CloudProjectManager
from lean.components.cloud.cloud_runner import CloudRunner
from lean.components.cloud.data_downloader import DataDownloader
from lean.components.cloud.module_manager import ModuleManager
from lean.components.cloud.pull_manager import PullManager
from lean.components.cloud.push_manager import PushManager
from lean.components.config.cli_config_manager import CLIConfigManager
from lean.components.config.optimizer_config_manager import OptimizerConfigManager
from lean.components.config.output_config_manager import OutputConfigManager
from lean.components.config.storage import Storage
from lean.components.docker.docker_manager import DockerManager
from lean.components.docker.lean_runner import LeanRunner
from lean.components.util.http_client import HTTPClient
from lean.components.util.library_manager import LibraryManager
from lean.components.util.logger import Logger
from lean.components.util.market_hours_database import MarketHoursDatabase
from lean.components.util.name_generator import NameGenerator
from lean.components.util.organization_manager import OrganizationManager
from lean.components.util.platform_manager import PlatformManager
from lean.components.util.project_manager import ProjectManager
from lean.components.util.task_manager import TaskManager
from lean.components.util.temp_manager import TempManager
from lean.components.util.update_manager import UpdateManager
from lean.components.util.xml_manager import XMLManager
from lean.constants import CACHE_PATH, CREDENTIALS_CONFIG_PATH, GENERAL_CONFIG_PATH


class Container:

    def __init__(self):
        self.initialize()

    def initialize(self,
                   docker_manager: Union[DockerManager, Any] = None,
                   api_client: Union[APIClient, Any] = None,
                   lean_runner: Union[LeanRunner, Any] = None,
                   cloud_runner: Union[CloudRunner, Any] = None,
                   push_manager: Union[PushManager, Any] = None,
                   organization_manager: Union[OrganizationManager, Any] = None,
                   project_config_manager: Union[ProjectConfigManager, Any] = None):
        """The Container class wires all reusable components together."""
        self.logger = Logger()

        self.platform_manager = PlatformManager()
        self.task_manager = TaskManager(self.logger)
        self.name_generator = NameGenerator()
        self.temp_manager = TempManager()
        self.xml_manager = XMLManager()
        self.http_client = HTTPClient(self.logger)

        self.general_storage = Storage(file=GENERAL_CONFIG_PATH)
        self.credentials_storage = Storage(file=CREDENTIALS_CONFIG_PATH)
        self.cache_storage = Storage(file=CACHE_PATH)

        self.cli_config_manager = CLIConfigManager(self.general_storage, self.credentials_storage)

        self.api_client = api_client
        if not self.api_client:
            self.api_client = APIClient(self.logger,
                                        self.http_client,
                                        user_id=self.cli_config_manager.user_id.get_value(),
                                        api_token=self.cli_config_manager.api_token.get_value())

        self.module_manager = ModuleManager(self.logger, self.api_client, self.http_client)

        self.project_config_manager = project_config_manager
        if not self.project_config_manager:
            self.project_config_manager = ProjectConfigManager(self.xml_manager)

        self.lean_config_manager = LeanConfigManager(self.logger,
                                                     self.cli_config_manager,
                                                     self.project_config_manager,
                                                     self.module_manager,
                                                     self.cache_storage)
        self.path_manager = PathManager(self.lean_config_manager, self.platform_manager)
        self.output_config_manager = OutputConfigManager(self.lean_config_manager)
        self.optimizer_config_manager = OptimizerConfigManager(self.logger)

        self.docker_manager = docker_manager
        if not self.docker_manager:
            self.docker_manager = DockerManager(self.logger, self.temp_manager, self.platform_manager)

        self.project_manager = ProjectManager(self.logger,
                                              self.project_config_manager,
                                              self.lean_config_manager,
                                              self.path_manager,
                                              self.xml_manager,
                                              self.platform_manager,
                                              self.cli_config_manager,
                                              self.docker_manager)
        self.library_manager = LibraryManager(self.logger,
                                              self.project_manager,
                                              self.project_config_manager,
                                              self.lean_config_manager,
                                              self.path_manager,
                                              self.xml_manager)

        self.organization_manager = organization_manager
        if not self.organization_manager:
            self.organization_manager = OrganizationManager(self.logger, self.lean_config_manager)

        self.cloud_runner = cloud_runner
        if not cloud_runner:
            self.cloud_runner = CloudRunner(self.logger, self.api_client, self.task_manager)
        self.pull_manager = PullManager(self.logger,
                                        self.api_client,
                                        self.project_manager,
                                        self.project_config_manager,
                                        self.library_manager,
                                        self.platform_manager,
                                        self.organization_manager)

        self.push_manager = push_manager
        if not push_manager:
            self.push_manager = PushManager(self.logger,
                                            self.api_client,
                                            self.project_manager,
                                            self.project_config_manager,
                                            self.organization_manager)
        self.data_downloader = DataDownloader(self.logger, self.api_client, self.lean_config_manager)
        self.cloud_project_manager = CloudProjectManager(self.api_client,
                                                         self.project_config_manager,
                                                         self.pull_manager,
                                                         self.push_manager,
                                                         self.path_manager,
                                                         self.project_manager,
                                                         self.organization_manager)

        self.lean_runner = lean_runner
        if not self.lean_runner:
            self.lean_runner = LeanRunner(self.logger,
                                          self.project_config_manager,
                                          self.lean_config_manager,
                                          self.output_config_manager,
                                          self.docker_manager,
                                          self.module_manager,
                                          self.project_manager,
                                          self.temp_manager,
                                          self.xml_manager)

        self.market_hours_database = MarketHoursDatabase(self.lean_config_manager)

        self.update_manager = UpdateManager(self.logger, self.http_client, self.cache_storage, self.docker_manager)


container = Container()
container.data_downloader.update_database_files()



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from click import command, argument

from lean.click import LeanCommand
from lean.container import container


@command(cls=LeanCommand)
@argument("key", type=str)
def delete(key: str) -> str:
    """
    Delete a value from the organization's cloud object store.

    :param key: The desired key name to delete.
    """
    organization_id = container.organization_manager.try_get_working_organization_id()
    api_client = container.api_client
    api_client.object_store.delete(key, organization_id)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Optional

from click import command, option, Choice, IntRange

from lean.click import DateParameter, LeanCommand, ensure_options
from lean.constants import DEFAULT_ENGINE_IMAGE
from lean.container import container


@command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@option("--start",
        type=DateParameter(),
        required=True,
        help="Start date for the data to generate in yyyyMMdd format")
@option("--end",
        type=DateParameter(),
        default=datetime.today().strftime("%Y%m%d"),
        help="End date for the data to generate in yyyyMMdd format (defaults to today)")
@option("--symbol-count",
        type=IntRange(min=0),
        help="The number of symbols to generate data for")
@option("--tickers",
        type=str,
        required=False,
        default="",
        help="Comma separated list of tickers to use for generated data")
@option("--security-type",
        type=Choice(["Equity", "Forex", "Cfd", "Future", "Crypto", "Option"], case_sensitive=False),
        default="Equity",
        help="The security type to generate data for (defaults to Equity)")
@option("--resolution",
        type=Choice(["Tick", "Second", "Minute", "Hour", "Daily"], case_sensitive=False),
        default="Minute",
        help="The resolution of the generated data (defaults to Minute)")
@option("--data-density",
        type=Choice(["Dense", "Sparse", "VerySparse"], case_sensitive=False),
        default="Dense",
        help="The density of the generated data (defaults to Dense)")
@option("--include-coarse",
        type=bool,
        default=True,
        help="Whether coarse universe data should be generated for Equity data (defaults to True)")
@option("--market",
        type=str,
        default="",
        help="The market to generate data for (defaults to standard market for the security type)")
@option("--quote-trade-ratio",
        type=float,
        default=1.0,
        help="The ratio of generated quotes to generated trades. Values larger than 1 mean more quotes than "
             "trades. Only used for Option, Future and Crypto (defaults to 1)")
@option("--random-seed",
        type=IntRange(min=0),
        default=None,
        help="The random number generator seed. Defaults to None, which means no seed will be used")
@option("--ipo-percentage",
        type=float,
        default=5.0,
        help="The probability each equity generated will have an IPO event. Note that this is not the total "
             "probability for all symbols generated. Only used for Equity (defaults to 5.0)")
@option("--rename-percentage",
        type=float,
        default=30.0,
        help="The probability each equity generated will have a rename event. Note that this is not the total "
             "probability for all symbols generated. Only used for Equity (defaults to 30.0)")
@option("--splits-percentage",
        type=float,
        default=15.0,
        help="The probability each equity generated will have a stock split event. Note that this is not the "
             "total probability for all symbols generated. Only used for Equity (defaults to 15.0)")
@option("--dividends-percentage",
        type=float,
        default=60.0,
        help="The probability each equity generated will have dividends. Note that this is not the probability "
             "for all symbols genearted. Only used for Equity (defaults to 60.0)")
@option("--dividend-every-quarter-percentage",
        type=float,
        default=30.0,
        help="The probability each equity generated will have a dividend event every quarter. Note that this is "
             "not the total probability for all symbols generated. Only used for Equity (defaults to 30.0)")
@option("--option-price-engine",
        type=str,
        default="BaroneAdesiWhaleyApproximationEngine",
        help="The stochastic process, and returns new pricing engine to run calculations for that option "
             "(defaults to BaroneAdesiWhaleyApproximationEngine)")
@option("--volatility-model-resolution",
        type=Choice(["Tick", "Second", "Minute", "Hour", "Daily"], case_sensitive=False),
        default="Daily",
        help="The volatility model period span (defaults to Daily)")
@option("--chain-symbol-count",
        type=IntRange(min=0),
        default=10,
        help="The size of the option chain (defaults to 10)")
@option("--image",
        type=str,
        help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@option("--update",
        is_flag=True,
        default=False,
        help="Pull the LEAN engine image before running the generator")
def generate(start: datetime,
             end: datetime,
             symbol_count: Optional[int],
             tickers: str,
             security_type: str,
             resolution: str,
             data_density: str,
             include_coarse: bool,
             market: str,
             quote_trade_ratio: float,
             random_seed: Optional[int],
             ipo_percentage: float,
             rename_percentage: float,
             splits_percentage: float,
             dividends_percentage: float,
             dividend_every_quarter_percentage: float,
             option_price_engine: str,
             volatility_model_resolution: str,
             chain_symbol_count: int,
             image: Optional[str],
             update: bool) -> None:
    """Generate random market data.

    \b
    This uses the random data generator in LEAN to generate realistic market data using a Brownian motion model.
    This generator supports the following security types, tick types and resolutions:
    | Security type | Generated tick types | Supported resolutions                |
    | ------------- | -------------------- | ------------------------------------ |
    | Equity        | Trade                | Tick, Second, Minute, Hour and Daily |
    | Forex         | Quote                | Tick, Second, Minute, Hour and Daily |
    | CFD           | Quote                | Tick, Second, Minute, Hour and Daily |
    | Future        | Trade and Quote      | Tick, Second, Minute, Hour and Daily |
    | Crypto        | Trade and Quote      | Tick, Second, Minute, Hour and Daily |
    | Option        | Trade and Quote      | Minute                               |

    \b
    The following data densities are available:
    - Dense: at least one data point per resolution step.
    - Sparse: at least one data point per 5 resolution steps.
    - VerySparse: at least one data point per 50 resolution steps.

    \b
    Example which generates minute data for 100 equity symbols since 2015-01-01:
    $ lean data generate --start=20150101 --symbol-count=100

    \b
    Example which generates daily data for 100 crypto symbols since 2015-01-01:
    $ lean data generate --start=20150101 --symbol-count=100 --security-type=Crypto --resolution=Daily

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    lean_config_manager = container.lean_config_manager
    data_dir = lean_config_manager.get_data_directory()

    entrypoint = ["dotnet", "QuantConnect.ToolBox.dll",
                  "--destination-dir", "/Lean/Data",
                  "--app", "randomdatagenerator",
                  "--start", start.strftime("%Y%m%d"),
                  "--end", end.strftime("%Y%m%d"),
                  "--security-type", security_type,
                  "--resolution", resolution,
                  "--data-density", data_density,
                  "--include-coarse", str(include_coarse).lower(),
                  "--market", market.lower(),
                  "--quote-trade-ratio", str(quote_trade_ratio),
                  "--ipo-percentage", str(ipo_percentage),
                  "--rename-percentage", str(rename_percentage),
                  "--splits-percentage", str(splits_percentage),
                  "--dividends-percentage", str(dividends_percentage),
                  "--dividend-every-quarter-percentage", str(dividend_every_quarter_percentage),
                  "--option-price-engine", option_price_engine,
                  "--volatility-model-resolution", volatility_model_resolution,
                  "--chain-symbol-count", str(chain_symbol_count)]

    if tickers:
        entrypoint.extend(["--tickers", tickers])
        entrypoint.extend(["--symbol-count", str(len(tickers.split(",")))])
    else:
        ensure_options(["symbol_count"])
        entrypoint.extend(["--symbol-count", str(symbol_count)])

    if random_seed:
        entrypoint.extend(["--random-seed", str(random_seed)])

    run_options = {
        "entrypoint": entrypoint,
        "volumes": {
            str(data_dir): {
                "bind": "/Lean/Data",
                "mode": "rw"
            }
        }
    }

    engine_image = container.cli_config_manager.get_engine_image(image)

    container.update_manager.pull_docker_image_if_necessary(engine_image, update)

    success = container.docker_manager.run_image(engine_image, **run_options)
    if not success:
        raise RuntimeError(
            "Something went wrong while running the random data generator, see the logs above for more information")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Optional

from click import prompt

from lean.models.logger import Option


class Logger:
    """The Logger class handles all output printing."""

    def __init__(self) -> None:
        """Creates a new Logger instance."""
        from rich.console import Console
        self._console = Console(markup=False, highlight=False, emoji=False, width=None)
        self.debug_logging_enabled = False

    def debug(self, message: Any) -> None:
        """Logs a debug message if debug logging is enabled.

        :param message: the message to log
        """
        if self.debug_logging_enabled:
            self._console.print(message, style="grey50")

    def info(self, message: Any) -> None:
        """Logs an info message.

        :param message: the message to log
        """
        self._console.print(message)

    def warn(self, message: Any) -> None:
        """Logs a warning message.

        :param message: the message to log
        """
        self._console.print(message, style="yellow")

    def error(self, message: Any) -> None:
        """Logs an error message.

        :param message: the message to log
        """
        self._console.print(message, style="red")

    def transient_progress(self, prefix: str = ""):
        """Creates a Progress instance.

        :param prefix: the text to show before the bar (defaults to a blank string)
        :return: a Progress instance which can be used to display progress bars
        """
        from rich.progress import BarColumn, Progress, TextColumn
        progress = Progress(TextColumn(prefix), BarColumn(),
                            console=self._console, transient=True)
        progress.start()
        return progress

    def progress(self, prefix: str = "", suffix: str = "{task.percentage:0.0f}%"):
        """Creates a Progress instance.

        :param prefix: the text to show before the bar (defaults to a blank string)
        :param suffix: the text to show after the bar (defaults to the task's percentage)
        :return: a Progress instance which can be used to display progress bars
        """
        from rich.progress import BarColumn, Progress, TextColumn
        progress = Progress(TextColumn(prefix), BarColumn(), TextColumn(suffix), console=self._console)
        progress.start()
        return progress

    def prompt_list(self, text: str, options: List[Option], default: Optional[str] = None, multiple: bool = False) -> Any:
        """Asks the user to select an option from a list of possible options.

        The user will not be prompted for input if there is only a single option.

        :param text: the text to display before prompting
        :param options: the available options
        :param default: the default value if no input is given
        :return: the chosen option's id
        """
        def validate_option(input_value: Any):
            try:
                index = int(input_value)
                if 0 < index <= len(options):
                    return options[index - 1].id
            except ValueError:
                option = next((option for option in options if option.label == input_value), None)
                if option is not None:
                    return option.id

            self.info("Please enter the number or label of an option")

        if len(options) == 1:
            self.info(f"{text}: {options[0].label}")
            return options[0].id

        self.info(f"{text}:")
        for i, option in enumerate(options):
            self.info(f"{i + 1}) {option.label}")

        while True:
            if not multiple:
                user_input = prompt("Enter an option", type=str, default=default, show_default=True)
                user_selected_value = validate_option(user_input)
                if user_selected_value is not None:
                    return user_selected_value
            else:
                user_selected_values = []
                user_inputs = prompt("To enter multiple options, separate them with comma", type=str, default=default, show_default=True)
                user_inputs = str(user_inputs).strip(",").split(",")
                expected_outputs = len(user_inputs)
                for user_input in user_inputs:
                    user_selected_value = validate_option(user_input)
                    if user_selected_value is not None:
                        user_selected_values.append(user_selected_value)
                if len(user_selected_values) == expected_outputs:
                    return user_selected_values

    def prompt_password(self, text: str, default: Optional[str] = None, hide_input: bool = True) -> str:
        """Asks the user for a string value while masking the given input.

        :param text: the text to display before prompting
        :param default: the default value if no input is given
        :param hide_input: whether to hide the input
        :return: the given input
        """
        from platform import uname
        from sys import stdin
        from maskpass import askpass

        if default is not None:
            text = f"{text} [{'*' * len(default)}]"

        # Masking does not work properly in WSL2 and when the input is not coming from a keyboard
        if "microsoft" in uname().release.lower() or not stdin.isatty() or not hide_input:
            return prompt(text, default=default, show_default=False, hide_input=hide_input)

        while True:
            user_input = askpass(f"{text}: ")

            if len(user_input) == 0 and default is not None:
                return default

            if len(user_input) > 0:
                return user_input



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, Dict, Optional, List

from lean.components.cloud.module_manager import ModuleManager
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.output_config_manager import OutputConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.docker.docker_manager import DockerManager
from lean.components.util.logger import Logger
from lean.components.util.project_manager import ProjectManager
from lean.components.util.temp_manager import TempManager
from lean.components.util.xml_manager import XMLManager
from lean.constants import MODULES_DIRECTORY, LEAN_ROOT_PATH, DEFAULT_DATA_DIRECTORY_NAME, \
    DEFAULT_LEAN_DOTNET_FRAMEWORK, DEFAULT_LEAN_PYTHON_VERSION
from lean.constants import DOCKER_PYTHON_SITE_PACKAGES_PATH
from lean.models.docker import DockerImage
from lean.models.utils import DebuggingMethod

class LeanRunner:
    """The LeanRunner class contains the code that runs the LEAN engine locally."""

    def __init__(self,
                 logger: Logger,
                 project_config_manager: ProjectConfigManager,
                 lean_config_manager: LeanConfigManager,
                 output_config_manager: OutputConfigManager,
                 docker_manager: DockerManager,
                 module_manager: ModuleManager,
                 project_manager: ProjectManager,
                 temp_manager: TempManager,
                 xml_manager: XMLManager) -> None:
        """Creates a new LeanRunner instance.

        :param logger: the logger that is used to print messages
        :param project_config_manager: the ProjectConfigManager instance to retrieve project configuration from
        :param lean_config_manager: the LeanConfigManager instance to retrieve Lean configuration from
        :param output_config_manager: the OutputConfigManager instance to retrieve backtest/live configuration from
        :param docker_manager: the DockerManager instance which is used to interact with Docker
        :param module_manager: the ModuleManager instance to retrieve the installed modules from
        :param project_manager: the ProjectManager instance to use for copying source code to output directories
        :param temp_manager: the TempManager instance to use for creating temporary directories
        :param xml_manager: the XMLManager instance to use for reading/writing XML files
        """
        self._logger = logger
        self._project_config_manager = project_config_manager
        self._lean_config_manager = lean_config_manager
        self._output_config_manager = output_config_manager
        self._docker_manager = docker_manager
        self._module_manager = module_manager
        self._project_manager = project_manager
        self._temp_manager = temp_manager
        self._xml_manager = xml_manager

    def run_lean(self,
                 lean_config: Dict[str, Any],
                 environment: str,
                 algorithm_file: Path,
                 output_dir: Path,
                 image: DockerImage,
                 debugging_method: Optional[DebuggingMethod],
                 release: bool,
                 detach: bool,
                 extra_docker_config: Optional[Dict[str, Any]] = None,
                 paths_to_mount: Optional[Dict[str, str]] = None) -> None:
        """Runs the LEAN engine locally in Docker.

        Raises an error if something goes wrong.

        :param lean_config: the LEAN configuration to use
        :param environment: the environment to run the algorithm in
        :param algorithm_file: the path to the file containing the algorithm
        :param output_dir: the directory to save output data to
        :param image: the LEAN engine image to use
        :param debugging_method: the debugging method if debugging needs to be enabled, None if not
        :param release: whether C# projects should be compiled in release configuration instead of debug
        :param detach: whether LEAN should run in a detached container
        :param extra_docker_config: additional docker configurations
        :param paths_to_mount: additional paths to mount to the container
        """
        self._logger.debug(f'LeanRunner().run_lean: lean_config: {lean_config}')
        project_dir = algorithm_file.parent

        # The dict containing all options passed to `docker run`
        # See all available options at https://docker-py.readthedocs.io/en/stable/containers.html
        run_options = self.get_basic_docker_config(lean_config,
                                                   algorithm_file,
                                                   output_dir,
                                                   debugging_method,
                                                   release,
                                                   detach,
                                                   image,
                                                   paths_to_mount)

        # Add known additional run options from the extra docker config
        self.parse_extra_docker_config(run_options, extra_docker_config)

        # Set up PTVSD debugging
        if debugging_method == DebuggingMethod.PTVSD:
            run_options["ports"]["5678"] = "5678"

        # Set up VSDBG debugging
        if debugging_method == DebuggingMethod.VSDBG:
            run_options["name"] = "lean_cli_vsdbg"

            # lean_cli_vsdbg is not unique, so we don't store the container name in the output directory's config
            output_config = self._output_config_manager.get_output_config(output_dir)
            output_config.delete("container")

        # Set up Rider debugging
        if debugging_method == DebuggingMethod.Rider:
            run_options["ports"]["22"] = "2222"
            run_options["commands"].append("/usr/sbin/enable_insecure_key")
            run_options["commands"].append("chmod 600 /etc/insecure_key")
            run_options["commands"].append('echo "HostKey /etc/insecure_key" >> /etc/ssh/sshd_config')
            run_options["commands"].append("/usr/sbin/sshd")

            self._docker_manager.create_volume("lean_cli_rider")
            run_options["volumes"]["lean_cli_rider"] = {
                "bind": "/root/.local/share/JetBrains",
                "mode": "rw"
            }

        # Setup debugging with QC Extension
        if debugging_method == DebuggingMethod.LocalPlatform:
            run_options["ports"]["5678"] = "0" # Using port 0 will assign a random port every time

        run_options["commands"].append("exec dotnet QuantConnect.Lean.Launcher.dll")

        # Copy the project's code to the output directory
        self._project_manager.copy_code(algorithm_file.parent, output_dir / "code")

        # Run the engine and log the result
        # Run as a subprocess to capture the output before logging it

        # Format error messages for cleaner output logs
        run_options["format_output"] = self.format_error_before_logging

        success = self._docker_manager.run_image(image, **run_options)

        cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        relative_project_dir = project_dir.relative_to(cli_root_dir)
        relative_output_dir = output_dir.relative_to(cli_root_dir)

        if debugging_method == DebuggingMethod.LocalPlatform:
            # set the container port mapped with host in config for later retrieval
            port = self._docker_manager.get_container_port(run_options["name"], "5678/tcp")
            output_config = self._output_config_manager.get_output_config(output_dir)
            output_config.set("debugport", port)

        if detach:
            self._temp_manager.delete_temporary_directories_when_done = False

            self._logger.info(
                f"Successfully started '{relative_project_dir}' in the '{environment}' environment in the '{run_options['name']}' container")
            self._logger.info(f"The output will be stored in '{relative_output_dir}'")
            self._logger.info("You can use Docker's own commands to manage the detached container")
        elif success:
            self._logger.info(
                f"Successfully ran '{relative_project_dir}' in the '{environment}' environment and stored the output in '{relative_output_dir}'")
        else:
            raise RuntimeError(
                f"Something went wrong while running '{relative_project_dir}' in the '{environment}' environment, the output is stored in '{relative_output_dir}'")

    def get_basic_docker_config(self,
                                lean_config: Dict[str, Any],
                                algorithm_file: Path,
                                output_dir: Path,
                                debugging_method: Optional[DebuggingMethod],
                                release: bool,
                                detach: bool,
                                image: DockerImage,
                                paths_to_mount: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
        """Creates a basic Docker config to run the engine with.

        This method constructs the parts of the Docker config that is the same for both the engine and the optimizer.

        :param lean_config: the LEAN configuration to use
        :param algorithm_file: the path to the file containing the algorithm
        :param output_dir: the directory to save output data to
        :param debugging_method: the debugging method if debugging needs to be enabled, None if not
        :param release: whether C# projects should be compiled in release configuration instead of debug
        :param detach: whether LEAN should run in a detached container
        :param image: The docker image that will be used
        :return: the Docker configuration containing basic configuration to run Lean
        :param paths_to_mount: additional paths to mount to the container
        """
        from docker.types import Mount
        from uuid import uuid4
        from json import dumps

        project_dir = algorithm_file.parent
        project_config = self._project_config_manager.get_project_config(project_dir)
        docker_project_config = project_config.get("docker", {})

        # Force the use of the LocalDisk map/factor providers if no recent zip present and not using ApiDataProvider
        data_dir = self._lean_config_manager.get_data_directory()
        if lean_config.get("data-provider", None) != "QuantConnect.Lean.Engine.DataFeeds.ApiDataProvider":
            self._force_disk_provider_if_necessary(lean_config,
                                                   "map-file-provider",
                                                   "QuantConnect.Data.Auxiliary.LocalZipMapFileProvider",
                                                   "QuantConnect.Data.Auxiliary.LocalDiskMapFileProvider",
                                                   data_dir / "equity" / "usa" / "map_files")
            self._force_disk_provider_if_necessary(lean_config,
                                                   "factor-file-provider",
                                                   "QuantConnect.Data.Auxiliary.LocalZipFactorFileProvider",
                                                   "QuantConnect.Data.Auxiliary.LocalDiskFactorFileProvider",
                                                   data_dir / "equity" / "usa" / "factor_files")

        # Create the output directory if it doesn't exist yet
        if not output_dir.exists():
            output_dir.mkdir(parents=True)

        # Create the storage directory if it doesn't exist yet
        storage_dir = self._lean_config_manager.get_cli_root_directory() / "storage"
        if not storage_dir.exists():
            storage_dir.mkdir(parents=True)

        lean_config["debug-mode"] = self._logger.debug_logging_enabled
        lean_config["data-folder"] = "/Lean/Data"
        lean_config["results-destination-folder"] = "/Results"
        lean_config["object-store-root"] = "/Storage"

        # The dict containing all options passed to `docker run`
        # See all available options at https://docker-py.readthedocs.io/en/stable/containers.html
        run_options: Dict[str, Any] = {
            "detach": detach,
            "commands": [],
            "environment": docker_project_config.get("environment", {}),
            "stop_signal": "SIGINT" if debugging_method is None else "SIGKILL",
            "mounts": [],
            "volumes": {},
            "ports": docker_project_config.get("ports", {})
        }

        # mount the paths passed in
        self.mount_paths(paths_to_mount, lean_config, run_options)

        # mount the project and library directories
        self.mount_project_and_library_directories(project_dir, run_options)

        # Mount the data directory
        run_options["volumes"][str(data_dir)] = {
            "bind": "/Lean/Data",
            "mode": "rw"
        }

        # Mount the output directory
        run_options["volumes"][str(output_dir)] = {
            "bind": "/Results",
            "mode": "rw"
        }

        # Mount the local object store directory
        run_options["volumes"][str(storage_dir)] = {
            "bind": "/Storage",
            "mode": "rw"
        }

        # Mount all local files referenced in the Lean config
        cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        files_to_mount = [
            ("transaction-log", cli_root_dir),
            ("terminal-link-symbol-map-file", cli_root_dir / DEFAULT_DATA_DIRECTORY_NAME / "symbol-properties")
        ]
        for key, base_path in files_to_mount:
            if key not in lean_config or lean_config[key] == "":
                continue

            lean_config_entry = Path(lean_config[key])
            local_path = lean_config_entry if lean_config_entry.is_absolute() else base_path / lean_config_entry
            if not local_path.exists():
                local_path.parent.mkdir(parents=True, exist_ok=True)
                local_path.touch()

            run_options["mounts"].append(Mount(target=f"/Files/{key}",
                                               source=str(local_path),
                                               type="bind",
                                               read_only=False))

            lean_config[key] = f"/Files/{key}"

        # Update all hosts that need to point to the host's localhost to host.docker.internal so they resolve properly
        for key in ["terminal-link-server-host"]:
            if key not in lean_config:
                continue

            if lean_config[key] == "localhost" or lean_config[key] == "127.0.0.1":
                lean_config[key] = "host.docker.internal"

        set_up_common_csharp_options_called = False

        # Set up modules
        installed_packages = self._module_manager.get_installed_packages()
        if len(installed_packages) > 0:
            self._logger.debug(f"LeanRunner.run_lean(): installed packages {len(installed_packages)}")
            self.set_up_common_csharp_options(run_options)
            set_up_common_csharp_options_called = True

            # Mount the modules directory
            run_options["volumes"][MODULES_DIRECTORY] = {
                "bind": "/Modules",
                "mode": "ro"
            }

            # Add the modules directory as a NuGet source root
            run_options["commands"].append("dotnet nuget add source /Modules")

            # Create a C# project used to resolve the dependencies of the modules
            run_options["commands"].append("mkdir /ModulesProject")
            run_options["commands"].append("dotnet new sln -o /ModulesProject")

            framework_ver = self._docker_manager.get_image_label(image, 'target_framework',
                                                                 DEFAULT_LEAN_DOTNET_FRAMEWORK)
            run_options["commands"].append(f"dotnet new classlib -o /ModulesProject -f {framework_ver} --no-restore")
            run_options["commands"].append("rm /ModulesProject/Class1.cs")

            # Add all modules to the project, automatically resolving all dependencies
            for package in installed_packages:
                self._logger.debug(f"LeanRunner.run_lean(): Adding module {package} to the project")
                run_options["commands"].append(f"rm -rf /root/.nuget/packages/{package.name.lower()}")
                run_options["commands"].append(
                    f"dotnet add /ModulesProject package {package.name} --version {package.version}")

            # Copy all module files to /Lean/Launcher/bin/Debug, but don't overwrite anything that already exists
            run_options["commands"].append(
                "python /copy_csharp_dependencies.py /Compile/obj/ModulesProject/project.assets.json")

        # Set up language-specific run options
        self.setup_language_specific_run_options(run_options, project_dir, algorithm_file,
                                                 set_up_common_csharp_options_called, release,
                                                 image)

        # Save the final Lean config to a temporary file so we can mount it into the container
        config_path = self._temp_manager.create_temporary_directory() / "config.json"
        with config_path.open("w+", encoding="utf-8") as file:
            file.write(dumps(lean_config, indent=4))

        # Mount the Lean config
        run_options["mounts"].append(Mount(target=f"{LEAN_ROOT_PATH}/config.json",
                                           source=str(config_path),
                                           type="bind",
                                           read_only=True))

        # Assign the container a name and store it in the output directory's configuration
        if "container-name" in lean_config:
            run_options["name"] = lean_config["container-name"]
        else:
            run_options["name"] = f"lean_cli_{str(uuid4()).replace('-', '')}"

        # set the hostname
        if "hostname" in lean_config:
            run_options["hostname"] = lean_config["hostname"]

        output_config = self._output_config_manager.get_output_config(output_dir)
        output_config.set("container", run_options["name"])
        if "backtest-name" in lean_config:
            output_config.set("backtest-name", lean_config["backtest-name"])
        if "environment" in lean_config and "environments" in lean_config:
            environment = lean_config["environments"][lean_config["environment"]]
            if "live-mode-brokerage" in environment:
                output_config.set("brokerage", environment["live-mode-brokerage"].split(".")[-1])

        return run_options

    def set_up_python_options(self, project_dir: Path, run_options: Dict[str, Any], image: DockerImage) -> None:
        """Sets up Docker run options specific to Python projects.

        :param project_dir: the path to the project directory
        :param run_options: the dictionary to append run options to
        :param image: the docker image that will be used
        """

        from docker.types import Mount
        # Compile python files
        source_files = self._project_manager.get_source_files(project_dir)
        source_files = [file.relative_to(
            project_dir).as_posix() for file in source_files]
        source_files = [f'"/LeanCLI/{file}"' for file in source_files]

        # Only need to compile files in backtest/live (where the files were mounted in "/LeanCLI") but not research
        run_options["commands"].append(
            f"""if [ -d '/LeanCLI' ];
            then
                python -m compileall {' '.join(source_files)};
            else
                echo '/LeanCLI is not mounted, skipping compilation...';
            fi""")

        # Combine the requirements from all library projects and the current project
        library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
        requirements_files = list(library_dir.rglob("requirements.txt")) + [project_dir / "requirements.txt"]
        requirements_files = [file for file in requirements_files if file.is_file()]
        requirements = self._concat_python_requirements(requirements_files)

        # Check if we have any dependencies to install, so we don't mount volumes needlessly
        if requirements == "":
            return

        # Create a requirements.txt file for the combined requirements
        requirements_txt = self._temp_manager.create_temporary_directory() / "requirements.txt"
        with requirements_txt.open("w+", encoding="utf-8") as file:
            file.write(requirements)

        # Mount the requirements.txt file
        run_options["mounts"].append(Mount(target="/requirements.txt",
                                           source=str(requirements_txt),
                                           type="bind",
                                           read_only=True))

        # Mount a volume to pip's cache directory so we only download packages once
        self._docker_manager.create_volume("lean_cli_pip")
        run_options["volumes"]["lean_cli_pip"] = {
            "bind": "/root/.cache/pip",
            "mode": "rw"
        }

        python_version = self._docker_manager.get_image_label(image, 'python_version',
                                                              DEFAULT_LEAN_PYTHON_VERSION)
        site_packages_path = DOCKER_PYTHON_SITE_PACKAGES_PATH.replace('{LEAN_PYTHON_VERSION}', python_version)

        # Mount a volume to the user packages directory so we don't install packages every time
        site_packages_volume = self._docker_manager.create_site_packages_volume(requirements_txt)
        run_options["volumes"][site_packages_volume] = {
            "bind": f"{site_packages_path}",
            "mode": "rw"
        }

        # Update PATH in the Docker container to prevent pip install warnings about its executables not being on PATH
        run_options["commands"].append('export PATH="$PATH:/root/.local/bin"')

        # Install custom libraries to the cached user packages directory
        # We only need to do this if it hasn't already been done before for this site packages volume
        # To keep track of this we create a special file in the site packages directory after installation
        # If this file already exists we can skip pip install completely
        marker_file = f"{site_packages_path}/pip-install-done"
        run_options["commands"].extend([
            f"! test -f {marker_file} && pip install --user --progress-bar off -r /requirements.txt",
            f"touch {marker_file}"
        ])

    def _concat_python_requirements(self, requirements_files: List[Path]) -> str:
        """Combines the requirements from multiple requirements.txt files.

        Ensures there are no duplicate requirements and that all output lines are valid.
        Requirements are sorted alphabetically to ensure consistent output.

        :param requirements_files: the paths to the requirements.txt files
        :return: the normalized requirements from all given requirements.txt files
        """
        from pkg_resources import Requirement
        requirements = []
        for file in requirements_files:
            for line in file.read_text(encoding="utf-8").splitlines():
                try:
                    requirements.append(Requirement.parse(line))
                except ValueError:
                    pass

        requirements = [str(requirement) for requirement in requirements]
        requirements = sorted(set(requirements))
        return "\n".join(requirements)

    def set_up_csharp_options(self, project_dir: Path, run_options: Dict[str, Any], release: bool,
                              image: DockerImage) -> None:
        """Sets up Docker run options specific to C# projects.

        :param project_dir: the path to the project directory
        :param run_options: the dictionary to append run options to
        :param release: whether C# projects should be compiled in release configuration instead of debug
        :param image: the docker image that will be used
        """
        compile_root = self._get_csharp_compile_root(project_dir)

        if compile_root != project_dir:
            # The /LeanCLI needs to be properly mounted to the compile root when the project is part of a solution
            run_options["volumes"].pop(str(project_dir))
            run_options["volumes"][str(compile_root)] = {
                "bind": "/LeanCLI",
                "mode": "rw"
            }

        # Ensure all .csproj files refer to the version of LEAN in the Docker container
        csproj_temp_dir = self._temp_manager.create_temporary_directory()
        for path in compile_root.rglob("*.csproj"):
            self._ensure_csproj_uses_correct_lean(compile_root, path, csproj_temp_dir, run_options)

        framework_ver = self._docker_manager.get_image_label(image, 'target_framework',
                                                             DEFAULT_LEAN_DOTNET_FRAMEWORK)
        # Set up the MSBuild properties
        msbuild_properties = {
            "Configuration": "Release" if release else "Debug",
            "Platform": "AnyCPU",
            "TargetFramework": framework_ver,
            "OutputPath": "/Compile/bin",
            "GenerateAssemblyInfo": "false",
            "GenerateTargetFrameworkAttribute": "false",
            "AppendTargetFrameworkToOutputPath": "false",
            "AutoGenerateBindingRedirects": "true",
            "GenerateBindingRedirectsOutputType": "true",
            "AutomaticallyUseReferenceAssemblyPackages": "false",
            "CopyLocalLockFileAssemblies": "true",
            "PathMap": f"/LeanCLI={str(compile_root)}",
            "NoWarn": ["MSB3277"]
        }

        # Find the .csproj file to compile
        project_file = next(project_dir.glob("*.csproj"))

        # Inherit NoWarn from the user's .csproj
        csproj = self._xml_manager.parse(project_file.read_text(encoding="utf-8"))
        existing_no_warn = csproj.find(".//NoWarn")
        from re import split
        if existing_no_warn is not None:
            codes = [c for c in split(r"[^a-zA-Z0-9]", existing_no_warn.text) if c != ""]
            msbuild_properties["NoWarn"] += codes

        # Turn the NoWarn property into a string
        # %3B is the encoded version of ";", because a raw ";" is seen as a separator between properties
        msbuild_properties["NoWarn"] = "%3B".join(msbuild_properties["NoWarn"])

        # Build the project before running LEAN
        relative_project_file = str(project_file.relative_to(compile_root)).replace("\\", "/")
        msbuild_properties = ";".join(f"{key}={value}" for key, value in msbuild_properties.items())
        run_options["commands"].append(f'dotnet build "/LeanCLI/{relative_project_file}" "-p:{msbuild_properties}"')

        # Copy over the algorithm DLL
        # Copy over the project reference DLLs'
        # Copy over all output DLLs that don't already exist in /Lean/Launcher/bin/Debug
        run_options["commands"].append(f"cp -R -n /Compile/bin/. {LEAN_ROOT_PATH}/")

        # Copy over all library DLLs that don't already exist in /Lean/Launcher/bin/Debug
        # CopyLocalLockFileAssemblies does not copy the OS-specific DLLs to the output directory
        # We therefore use a custom Python script that does take the OS into account when deciding what to copy
        run_options["commands"].append(
            f'python /copy_csharp_dependencies.py "/Compile/obj/{project_file.stem}/project.assets.json"')

    def set_up_common_csharp_options(self, run_options: Dict[str, Any]) -> None:
        """Sets up common Docker run options that is needed for all C# work.

        This method is only called if the user has installed modules and/or if the project to run is written in C#.

        :param run_options: the dictionary to append run options to
        """
        from docker.types import Mount
        # Mount a volume to NuGet's cache directory so we only download packages once
        self._docker_manager.create_volume("lean_cli_nuget")
        run_options["volumes"]["lean_cli_nuget"] = {
            "bind": "/root/.nuget/packages",
            "mode": "rw"
        }

        # Reduce the dotnet output
        run_options["environment"]["DOTNET_NOLOGO"] = "true"
        run_options["environment"]["DOTNET_CLI_TELEMETRY_OPTOUT"] = "true"

        temp_files_directory = self._temp_manager.create_temporary_directory()

        # Create a Directory.Build.props file to ensure /Compile/obj is used instead of <project>/obj
        # This is necessary because the files in obj/ will contain absolute paths, which are valid only in the container
        # Rider is okay with this, but Visual Studio throws a lot of errors if this happens
        directory_build_props = temp_files_directory / "Directory.Build.props"
        with directory_build_props.open("w+", encoding="utf-8") as file:
            file.write("""
<Project>
    <PropertyGroup>
        <BaseIntermediateOutputPath>/Compile/obj/$(MSBuildProjectName)/</BaseIntermediateOutputPath>
        <IntermediateOutputPath>/Compile/obj/$(MSBuildProjectName)/</IntermediateOutputPath>
        <DefaultItemExcludes>$(DefaultItemExcludes);backtests/*/code/**;live/*/code/**;optimizations/*/code/**</DefaultItemExcludes>
        <NoWarn>CS0618</NoWarn>
    </PropertyGroup>
</Project>
            """.strip())

        # Create a Python script that can be used to copy the right C# dependencies to /Lean/Launcher/bin/Debug
        # This script copies the correct DLLs even if a project has OS-specific DLLs
        copy_csharp_dependencies = temp_files_directory / "copy_csharp_dependencies.py"
        with copy_csharp_dependencies.open("w+", encoding="utf-8") as file:
            file.write("""
import json
import os
import platform
import shutil
import sys
from pathlib import Path

project_assets = json.loads(Path(sys.argv[-1]).read_text(encoding="utf-8"))
package_folders = [Path(folder) for folder in project_assets["packageFolders"].keys()]

ubuntu_version = os.popen("lsb_release -rs").read().strip()
accepted_runtime_identifiers = ["base", "unix", "linux", "debian", "ubuntu", f"ubuntu.{ubuntu_version}"]

if platform.machine() in ["arm64", "aarch64"]:
    accepted_runtime_identifiers.extend(["unix-arm", "linux-arm", "debian-arm", "ubuntu-arm", f"ubuntu.{ubuntu_version}-arm"])
else:
    accepted_runtime_identifiers.extend(["unix-x64", "linux-x64", "debian-x64", "ubuntu-x64", f"ubuntu.{ubuntu_version}-x64"])

def copy_file(library_id, partial_path, file_data):
    if "rid" in file_data and file_data["rid"] not in accepted_runtime_identifiers:
        return

    if not file_data.get("copyToOutput", True):
        return

    for folder in package_folders:
        full_path = folder / library_id.lower() / partial_path
        if full_path.exists():
            break
    else:
        return

    output_name = file_data.get("outputPath", full_path.name)

    target_path = Path("/Lean/Launcher/bin/Debug") / output_name
    if not target_path.exists():
        target_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(full_path, target_path)

project_target = list(project_assets["targets"].keys())[0]
for library_id, library_data in project_assets["targets"][project_target].items():
    for key, value in library_data.get("runtimeTargets", {}).items():
        copy_file(library_id, key, value)

    for key, value in library_data.get("runtime", {}).items():
        copy_file(library_id, key, value)

    for key, value in library_data.get("contentFiles", {}).items():
        copy_file(library_id, key, value)
            """.strip())

        run_options["mounts"].extend([Mount(target="/Directory.Build.props",
                                            source=str(directory_build_props),
                                            type="bind",
                                            read_only=False),
                                      Mount(target="/copy_csharp_dependencies.py",
                                            source=str(copy_csharp_dependencies),
                                            type="bind",
                                            read_only=False)])

    def _get_csharp_compile_root(self, project_dir: Path) -> Path:
        """Returns the path to the directory that should be mounted to compile the project directory.

        If the project is part of a solution this is the solution root.
        If the project is not part of a solution this is the project directory itself.

        :param project_dir: the path to the project directory
        :return: the path that should be mounted in the Docker container when compiling the C# project
        """
        current_dir = project_dir.parent
        while current_dir.parent != current_dir:
            if next(current_dir.glob("*.sln"), None) is not None:
                return current_dir

            current_dir = current_dir.parent

        return project_dir

    def _ensure_csproj_uses_correct_lean(self,
                                         compile_root: Path,
                                         csproj_path: Path,
                                         temp_dir: Path,
                                         run_options: Dict[str, Any]) -> None:
        """Ensures a C# project is compiled using the version of LEAN in the Docker container.

        When a .csproj file refers to the NuGet version of LEAN,
        we mount a temporary file on top of it which refers to the version of LEAN in the Docker container.
        In the case there is a breaking change between the two,
        this causes compiling to fail with readable error messages instead of ugly messages at runtime.

        :param compile_root: the path that is mounted in the Docker container
        :param csproj_path: the path to the .csproj file
        :param temp_dir: the temporary directory in which temporary .csproj files should be placed
        :param run_options: the dictionary to append run options to
        """
        from docker.types import Mount
        csproj = self._xml_manager.parse(csproj_path.read_text(encoding="utf-8"))
        include_added = False

        for package_reference in csproj.iter("PackageReference"):
            if not package_reference.get("Include", "").lower().startswith("quantconnect."):
                continue

            if include_added:
                package_reference.getparent().remove(package_reference)

            package_reference.clear()

            package_reference.tag = "Reference"
            package_reference.set("Include", f"{LEAN_ROOT_PATH}/*.dll")
            package_reference.append(self._xml_manager.parse("<Private>False</Private>"))

            include_added = True

        if not include_added:
            return

        new_csproj_file = temp_dir / csproj_path.relative_to(compile_root)
        new_csproj_file.parent.mkdir(parents=True, exist_ok=True)
        with new_csproj_file.open("w+", encoding="utf-8") as file:
            file.write(self._xml_manager.to_string(csproj))

        run_options["mounts"].append(Mount(target=f"/LeanCLI/{csproj_path.relative_to(compile_root).as_posix()}",
                                           source=str(new_csproj_file),
                                           type="bind",
                                           read_only=True))

    def _force_disk_provider_if_necessary(self,
                                          lean_config: Dict[str, Any],
                                          config_key: str,
                                          zip_provider: str,
                                          disk_provider: str,
                                          zip_dir: Path) -> None:
        """Updates the Lean config to use the disk provider instead of the zip one if there are no zips to use.

        :param lean_config: the Lean config to update
        :param config_key: the key of the configuration property
        :param zip_provider: the fully classified name of the zip provider for this property
        :param disk_provider: the fully classified name of the disk provider for this property
        :param zip_dir: the directory where the zip provider looks for zip files
        """
        from re import sub
        from datetime import datetime

        if lean_config.get(config_key, None) != zip_provider:
            return

        if not zip_dir.exists():
            lean_config[config_key] = disk_provider
            return

        zip_names = sorted([f.name for f in zip_dir.iterdir() if f.name.endswith(".zip")], reverse=True)
        zip_names = [sub(r"[^\d]", "", name) for name in zip_names]

        if len(zip_names) == 0 or (datetime.now() - datetime.strptime(zip_names[0], "%Y%m%d")).days > 7:
            lean_config[config_key] = disk_provider

    def setup_language_specific_run_options(self, run_options, project_dir, algorithm_file,
                                            set_up_common_csharp_options_called, release, image: DockerImage) -> None:
        # Set up language-specific run options
        if algorithm_file.name.endswith(".py"):
            self.set_up_python_options(project_dir, run_options, image)
        else:
            if not set_up_common_csharp_options_called:
                self.set_up_common_csharp_options(run_options)
            self.set_up_csharp_options(project_dir, run_options, release, image)

    def format_error_before_logging(self, chunk: str):
        from lean.components.util import compiler
        from json import loads
        errors = []

        # As we don't have algorithm type information. We can check errors for both
        # Python
        jsonString = compiler.get_errors("python", chunk)
        jsonData = loads(jsonString)
        errors.extend(jsonData["aErrors"])
        # CSharp
        jsonString = compiler.get_errors("csharp", chunk)
        jsonData = loads(jsonString)
        errors.extend(jsonData["aErrors"])

        for error in errors:
            self._logger.info(error)

    def mount_project_and_library_directories(self, project_dir: Path, run_options: Dict[str, Any]) -> None:
        # Mount the project directory
        run_options["volumes"][str(project_dir)] = {
            "bind": "/LeanCLI",
            "mode": "rw"
        }

        # Check if the user has library projects and mount the Library directory
        library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
        if library_dir.is_dir():
            # Mount the library projects
            run_options["volumes"][str(library_dir)] = {
                "bind": "/Library",
                "mode": "rw"
            }

    def mount_paths(self, paths_to_mount, lean_config, run_options):
        if not paths_to_mount:
            return

        from docker.types import Mount

        environment = {}
        if "environment" in lean_config and "environments" in lean_config:
            environment = lean_config["environments"][lean_config["environment"]]

        mounts = run_options["mounts"]

        for key, pathStr in paths_to_mount.items():
            path = Path(pathStr).resolve()
            target = f"/Files/{Path(path).name}"

            self._logger.info(f"Mounting {path} to {target}")

            mounts.append(Mount(target=target,
                                source=str(path),
                                type="bind",
                                read_only=True))
            environment[key] = target

    @staticmethod
    def parse_extra_docker_config(run_options: Dict[str, Any], extra_docker_config: Optional[Dict[str, Any]]) -> None:
        from docker.types import DeviceRequest
        # Add known additional run options from the extra docker config.
        # For now, only device_requests is supported
        if extra_docker_config is not None:
            if "device_requests" in extra_docker_config:
                run_options["device_requests"] = [DeviceRequest(**device_request)
                                                  for device_request in extra_docker_config["device_requests"]]

            if "volumes" in extra_docker_config:
                volumes = run_options.get("volumes")
                if not volumes:
                    volumes = run_options["volumes"] = {}
                volumes.update(extra_docker_config["volumes"])



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def rename_internal_config_to_user_friendly_format(key: str) -> str:
    """
    Function to rename a string if it matches a specific key.

    Parameters:
    key (str): The input string.

    Returns:
    str: The renamed string if it matches the specific key and passes the validation,
         otherwise returns the original string.
    """
    if key is None or len(key) == 0:
        raise ValueError("Input string is null or empty")
    
    # Check if the input string matches the specific key
    if key == "data-queue-handler":
        return "data-provider-live"
    else:
        return key

    


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

from lean.constants import MODULE_BROKERAGE, MODULE_TYPE, MODULE_CLOUD_PLATFORM, MODULE_PLATFORM, \
    MODULE_DATA_QUEUE_HANDLER
from lean.models import json_modules
from lean.models.json_module import JsonModule

# load the modules
cloud_brokerages: List[JsonModule] = []
cloud_data_queue_handlers: List[JsonModule] = []

for json_module in json_modules:
    module_type = json_module[MODULE_TYPE]
    platform = json_module[MODULE_PLATFORM]

    if MODULE_CLOUD_PLATFORM in platform:
        if MODULE_BROKERAGE in module_type:
            cloud_brokerages.append(JsonModule(json_module, MODULE_BROKERAGE, MODULE_CLOUD_PLATFORM))
        if MODULE_DATA_QUEUE_HANDLER in module_type:
            cloud_data_queue_handlers.append(JsonModule(json_module, MODULE_DATA_QUEUE_HANDLER, MODULE_CLOUD_PLATFORM))



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from click import command, Group

from lean.models.errors import MoreInfoError


class AnyCommandGroup(Group):
    """A command group that can be used to have single action to perform on any command or sub-command in the group,
    regardless of the sub-command name.

    For instance, the following would raise an exception on any of these calls:
    `mycommand`, `mycommand subcommand`, `mycommand anothersubcommand`, etc.

    ```
@command(cls=AnyCommandGroup, invoke_without_command=True)
def mycommand() -> None:
    raise MoreInfoError("Message")
    ```
    """

    def get_command(self, ctx, cmd_name):
        # always return the same command for any sub-command
        return self

    def resolve_command(self, ctx, args):
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


@command(cls=AnyCommandGroup, invoke_without_command=True)
def gui() -> None:
    """Work with the local GUI."""
    # This method is only used to warn users about GUI deprecation and point them to the new VSCode extension.
    raise MoreInfoError(
        "The LEAN GUI was deprecated in 2019 and has been replaced by a VSCode extension "
        "which allows seamless local and cloud quantitative research. "
        "See more information in the VSCode Marketplace.",
        "https://marketplace.visualstudio.com/items?itemName=quantconnect.quantconnect")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Dict, Any

from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.util.logger import Logger


def get_organization(config: Dict[str, Any]):
    organization_id = None
    if "job-organization-id" in config:
        organization_id = config["job-organization-id"]
    elif "organization-id" in config:
        # for backwards compatibility with local platform and old lean cli setups
        organization_id = config["organization-id"]
    return organization_id


class OrganizationManager:
    """The OrganizationManager class provides utilities to handle the working organization."""

    def __init__(self, logger: Logger, lean_config_manager: LeanConfigManager) -> None:
        """Creates a new OrganizationManager instance.

        :param logger: the logger to use to log messages with
        :param lean_config_manager: the LeanConfigManager to use to manipulate the lean configuration file
        """
        self._logger = logger
        self._lean_config_manager = lean_config_manager

        self._working_organization_id = None

    def get_working_organization_id(self) -> Optional[str]:
        """Gets the id of the working organization in the current Lean CLI directory.

        :return: the id of the working organization. None if the organization id was not found in the lean config
        """
        if self._working_organization_id is None:
            lean_config = self._lean_config_manager.get_lean_config()
            self._working_organization_id = get_organization(lean_config)

        return self._working_organization_id

    def try_get_working_organization_id(self) -> Optional[str]:
        """Gets the id of the working organization in the current Lean CLI directory.

        :return: the id of the working organization
        :raises RuntimeError: if the working organization is not found in the lean config
        """
        organization_id = self.get_working_organization_id()

        if organization_id is None:
            raise RuntimeError(
                "The working organization for this Lean CLI folder could not be determined.\n"
                "Make sure you run `lean init` on an empty folder for each organization you are a member of")

        return organization_id

    def configure_working_organization_id(self, organization_id: str) -> None:
        """
        Configures the working organization id in the Lean config

        :param organization_id: the working organization di
        """
        self._lean_config_manager.set_properties({"job-organization-id": organization_id})
        # for backwards compatibility with local platform
        self._lean_config_manager.set_properties({"organization-id": organization_id})
        self._working_organization_id = organization_id



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from lean.models.pydantic import WrappedBaseModel, validator


# The models in this module are all parts of responses from the QuantConnect API
# The keys of properties are not changed, so they don't obey the rest of the project's naming conventions


class ProjectEncryptionKey(WrappedBaseModel):
    id: str
    name: str

class QCCollaborator(WrappedBaseModel):
    uid: int
    liveControl: bool
    permission: str
    profileImage: str
    name: str
    owner: bool = False


class QCParameter(WrappedBaseModel):
    key: str
    value: str
    min: Optional[float]
    max: Optional[float]
    step: Optional[float]
    type: Optional[str]


class QCLanguage(str, Enum):
    CSharp = "C#"
    FSharp = "F#"
    VisualBasic = "VB"
    Java = "Ja"
    Python = "Py"


class QCProjectLibrary(WrappedBaseModel):
    projectId: int
    libraryName: str
    ownerName: str
    access: bool

    def __hash__(self):
        return hash(self.projectId)

    def __eq__(self, other: Any):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.projectId == other.projectId


class QCProject(WrappedBaseModel):
    projectId: int
    organizationId: str
    name: str
    description: str
    modified: datetime
    created: datetime
    language: QCLanguage
    collaborators: List[QCCollaborator]
    leanVersionId: int
    leanPinnedToMaster: bool
    leanEnvironment: int
    parameters: List[QCParameter]
    libraries: List[QCProjectLibrary]
    encrypted: Optional[bool] = False
    encryptionKey: Optional[ProjectEncryptionKey] = None

    @validator("parameters", pre=True)
    def process_parameters_dict(cls, value: Any) -> Any:
        if isinstance(value, dict):
            return list(value.values())
        return value

    def get_url(self) -> str:
        """Returns the url of the project page in the cloud.

        :return: a url which when visited opens an Algorithm Lab tab containing the project
        """
        return f"https://www.quantconnect.com/project/{self.projectId}"

    def __hash__(self):
        return hash(self.projectId)

    def __eq__(self, other: Any):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.projectId == other.projectId


class QCCreatedProject(WrappedBaseModel):
    projectId: int
    name: str
    modified: datetime
    created: datetime


class QCFullFile(WrappedBaseModel):
    name: str
    content: str
    modified: datetime
    isLibrary: bool


class QCMinimalFile(WrappedBaseModel):
    name: str
    content: str
    modified: datetime


class QCCompileState(str, Enum):
    InQueue = "InQueue"
    BuildSuccess = "BuildSuccess"
    BuildError = "BuildError"


class QCCompile(WrappedBaseModel):
    compileId: str
    state: QCCompileState


class QCCompileWithLogs(QCCompile):
    logs: List[str]


class QCBacktest(WrappedBaseModel):
    backtestId: str
    projectId: int
    status: str
    name: str
    note: Optional[str] = None
    created: datetime
    completed: bool
    progress: float
    result: Optional[Any] = None
    error: Optional[str] = None
    stacktrace: Optional[str] = None
    runtimeStatistics: Optional[Dict[str, str]]
    statistics: Optional[Union[Dict[str, str], List[Any]]]
    totalPerformance: Optional[Any]

    def is_complete(self) -> bool:
        """Returns whether the backtest has completed in the cloud.

        :return: True if the backtest is complete, False if not
        """
        if self.error is not None:
            return True

        if not self.completed:
            return False

        has_runtime_statistics = self.runtimeStatistics is not None
        has_statistics = self.statistics is not None and not isinstance(self.statistics, list)
        return has_runtime_statistics and has_statistics

    def get_url(self) -> str:
        """Returns the url of the backtests results page in the cloud.

        :return: a url which when visited opens an Algorithm Lab tab containing the backtest's results
        """
        return f"https://www.quantconnect.com/project/{self.projectId}/{self.backtestId}"

    def get_statistics_table(self):
        """Converts the statistics into a pretty table.

        :return: a table containing all statistics
        """
        from rich import box
        from rich.table import Table
        from rich.text import Text

        stats = []

        for key, value in self.runtimeStatistics.items():
            stats.append(key)

            if "-" in value:
                stats.append(Text.from_markup(f"[red]{value}[/red]"))
            elif any(char.isdigit() and int(char) > 0 for char in value):
                stats.append(Text.from_markup(f"[green]{value}[/green]"))
            else:
                stats.append(value)

        if len(stats) % 4 != 0:
            stats.extend(["", ""])

        end_of_first_section = len(stats)

        for key, value in self.statistics.items():
            stats.extend([key, value])

        if len(stats) % 4 != 0:
            stats.extend(["", ""])

        table = Table(box=box.SQUARE)
        table.add_column("Statistic", overflow="fold")
        table.add_column("Value", overflow="fold")
        table.add_column("Statistic", overflow="fold")
        table.add_column("Value", overflow="fold")

        for i in range(int(len(stats) / 4)):
            start = i * 4
            end = (i + 1) * 4
            table.add_row(*stats[start:end], end_section=end_of_first_section == end)

        return table


class QCNodePrice(WrappedBaseModel):
    monthly: int
    yearly: int


class QCNode(WrappedBaseModel):
    id: str
    name: str
    projectName: str
    description: str
    usedBy: str
    sku: str
    busy: bool
    price: QCNodePrice
    speed: float
    cpu: int
    ram: float
    assets: int
    host: Optional[str]


class QCNodeList(WrappedBaseModel):
    backtest: List[QCNode]
    research: List[QCNode]
    live: List[QCNode]


class QCLiveAlgorithmStatus(str, Enum):
    DeployError = "DeployError"
    InQueue = "InQueue"
    Running = "Running"
    Stopped = "Stopped"
    Liquidated = "Liquidated"
    Deleted = "Deleted"
    Completed = "Completed"
    RuntimeError = "RuntimeError"
    Invalid = "Invalid"
    LoggingIn = "LoggingIn"
    Initializing = "Initializing"
    History = "History"


class QCRestResponse(WrappedBaseModel):
    success: bool
    error: Optional[List[str]]

class QCMinimalLiveAlgorithm(WrappedBaseModel):
    projectId: int
    deployId: str
    status: Optional[QCLiveAlgorithmStatus] = None

    def get_url(self) -> str:
        """Returns the url of the live deployment in the cloud.

        :return: a url which when visited opens an Algorithm Lab tab containing the live deployment
        """
        return f"https://www.quantconnect.com/project/{self.projectId}/live"


class QCFullLiveAlgorithm(QCMinimalLiveAlgorithm):
    projectId: int
    deployId: str
    status: Optional[QCLiveAlgorithmStatus] = None
    launched: datetime
    stopped: Optional[datetime]
    brokerage: str
    subscription: str
    error: str


class QCEmailNotificationMethod(WrappedBaseModel):
    address: str
    subject: str


class QCWebhookNotificationMethod(WrappedBaseModel):
    address: str
    headers: Dict[str, str]


class QCSMSNotificationMethod(WrappedBaseModel):
    phoneNumber: str


class QCTelegramNotificationMethod(WrappedBaseModel):
    id: str
    token: Optional[str] = None


QCNotificationMethod = Union[QCEmailNotificationMethod, QCWebhookNotificationMethod, QCSMSNotificationMethod, QCTelegramNotificationMethod]


class QCCard(WrappedBaseModel):
    brand: str
    expiration: str
    last4: str


class QCAccount(WrappedBaseModel):
    organizationId: str
    card: Optional[QCCard] = None

    # Balance in QCC
    creditBalance: float


class QCOrganizationCreditMovement(WrappedBaseModel):
    date: str
    description: str
    type: str
    subtype: str
    amount: float

    # Balance in QCC
    balance: float


class QCOrganizationCredit(WrappedBaseModel):
    movements: List[QCOrganizationCreditMovement]

    # Balance in QCC
    balance: float


class QCOrganizationProductItem(WrappedBaseModel):
    productId: int
    name: str
    quantity: int
    unitPrice: float
    total: float


class QCOrganizationProduct(WrappedBaseModel):
    name: str
    items: List[QCOrganizationProductItem]


class QCOrganizationData(WrappedBaseModel):
    signedTime: Optional[int]
    current: bool


class QCOrganizationMember(WrappedBaseModel):
    id: int
    name: str
    isAdmin: bool
    email: str


class QCFullOrganization(WrappedBaseModel):
    id: str
    name: str
    seats: int
    type: str
    credit: QCOrganizationCredit
    products: List[QCOrganizationProduct]
    data: QCOrganizationData
    members: List[QCOrganizationMember]

    def has_security_master_subscription(self, id: int) -> bool:
        """Returns whether this organization has the Security Master subscription of a given Id

        :param id: the Id of the Security Master Subscription
        :return: True if the organization has a Security Master subscription, False if not
        """

        data_products_product = next((x for x in self.products if x.name == "Data"), None)
        if data_products_product is None:
            return False

        return any(x.productId == id for x in data_products_product.items)


class QCMinimalOrganization(WrappedBaseModel):
    id: str
    name: str
    type: str
    ownerName: str
    members: int
    preferred: bool


class QCSecurityType(str, Enum):
    Equity = "Equity"
    Forex = "Forex"
    CFD = "Cfd"
    Future = "Future"
    Crypto = "Crypto"
    Option = "Option"
    FutureOption = "FutureOption"


class QCResolution(str, Enum):
    Tick = "Tick"
    Second = "Second"
    Minute = "Minute"
    Hour = "Hour"
    Daily = "Daily"

    @classmethod
    def by_name(cls, name: str) -> 'QCResolution':
        """Returns the enum member with the same name as the given one, case insensitively.

        :param name: the name of the enum member (case insensitive)
        :return: the matching enum member
        """
        for k, v in cls.__members__.items():
            if k.lower() == name.lower():
                return v
        raise ValueError(f"QCResolution has no member named '{name}'")


class QCLink(WrappedBaseModel):
    link: str


class QCOptimizationBacktest(WrappedBaseModel):
    id: str
    name: str
    exitCode: int
    parameterSet: Dict[str, str]
    statistics: List[float] = []


class QCOptimization(WrappedBaseModel):
    optimizationId: str
    projectId: int
    status: str
    name: str
    backtests: Dict[str, QCOptimizationBacktest] = {}
    runtimeStatistics: Dict[str, str] = {}

    @validator("backtests", "runtimeStatistics", pre=True)
    def parse_empty_lists(cls, value: Any) -> Any:
        # If these fields have no data, they are assigned an array by default
        # For consistency we convert those empty arrays to empty dicts
        if isinstance(value, list):
            return {}
        return value

    def get_progress(self) -> float:
        """Returns the progress of the optimization between 0.0 and 1.0.

        :return: 0.0 if the optimization is 0% done, 1.0 if the optimization is 100% done, or somewhere in between
        """
        stats = self.runtimeStatistics
        if "Completed" in stats and "Failed" in stats and "Total" in stats:
            finished_backtests = float(stats["Completed"]) + float(stats["Failed"])
            total_backtests = float(stats["Total"])
            return finished_backtests / total_backtests
        return 0.0


class QCOptimizationEstimate(WrappedBaseModel):
    estimateId: str
    time: int
    balance: int


class QCDataVendor(WrappedBaseModel):
    vendorName: str
    regex: Any

    # Price in QCC
    price: Optional[float]

    @validator("regex", pre=True)
    def parse_regex(cls, value: Any) -> Any:
        from re import compile
        if isinstance(value, str):
            return compile(value[value.index("/") + 1:value.rindex("/")])
        return value


class QCDataInformation(WrappedBaseModel):
    datasources: Dict[str, Any]
    prices: List[QCDataVendor]
    agreement: str


class QCDatasetDelivery(str, Enum):
    CloudOnly = "cloud only"
    DownloadOnly = "download only"
    CloudAndDownload = "cloud & download"


class QCDatasetTag(WrappedBaseModel):
    name: str


class QCDataset(WrappedBaseModel):
    id: int
    name: str
    delivery: QCDatasetDelivery
    vendorName: str
    tags: List[QCDatasetTag]
    pending: bool


class QCUser(WrappedBaseModel):
    name: str
    profile: str
    badge: Optional[str]


class QCTerminalNewsItem(WrappedBaseModel):
    id: int
    type: str
    category: str
    title: str
    content: str
    image: str
    link: str
    year_deleted: Optional[Any]
    week_deleted: Optional[Any]
    created: datetime
    date: datetime


class QCLeanEnvironment(WrappedBaseModel):
    id: int
    name: str
    path: Optional[str]
    description: str
    public: bool



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from click import argument
from lean.commands.cloud.object_store import object_store
from lean.click import LeanCommand
from lean.container import container


@object_store.command(cls=LeanCommand, name="list", aliases=["ls"])
@argument("key", type=str, default="/")
def list(key: str):
    """
    List all values for the given root key in the organization's cloud object store.

    :param key: The desired root key to list.
    """
    organization_id = container.organization_manager.try_get_working_organization_id()
    api_client = container.api_client
    logger = container.logger
    data = api_client.object_store.list(key, organization_id)

    try:
        headers = ["key", "size", "folder", "name"]
        display_headers = ["Key", "Bytes", "Folder", "Filename"]
        rows = [[str(obj.get(header, "")) for header in headers] for obj in data['objects']]
        # sort rows by key
        rows.sort(key=lambda x: x[0])
        all_rows = [display_headers] + rows
        column_widths = [max(len(row[i]) for row in all_rows) for i in range(len(all_rows[0]))]
        for row in all_rows:
            logger.info("  ".join(value.ljust(width) for value, width in zip(row, column_widths)))
    except KeyError as e:
        logger.error(f"Key {key} not found.")
    except Exception as e:
        logger.error(f"Error: {e}")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional
from click import command, option, argument

from lean.click import LeanCommand, PathParameter
from lean.container import container


@command(cls=LeanCommand)
@argument("project", type=PathParameter(exists=True, file_okay=False, dir_okay=True))
@option("--key",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help="Path to the encryption key to use")
def encrypt(project: Path,
            key: Optional[Path]) -> None:
    """Encrypt your local project using the specified encryption key."""

    logger = container.logger
    project_manager = container.project_manager
    project_config_manager = container.project_config_manager
    project_config = project_config_manager.get_project_config(project)

    # Check if the project is already encrypted
    if project_config.get('encrypted', False):
        logger.info(f"Local files encrypted successfully.")
        return
    
    encryption_key: Path = project_config.get('encryption-key-path', None)
    from lean.components.util.encryption_helper import get_and_validate_user_input_encryption_key
    encryption_key = get_and_validate_user_input_encryption_key(key, encryption_key)

    organization_id = container.organization_manager.try_get_working_organization_id()

    source_files = project_manager.get_source_files(project)
    try:
        from lean.components.util.encryption_helper import get_encrypted_file_content_for_local_project
        encrypted_data = get_encrypted_file_content_for_local_project(project, 
                            source_files, encryption_key, project_config_manager, organization_id)
    except Exception as e:
        raise RuntimeError(f"Could not encrypt project {project}: {e}")
    for file, encrypted in zip(source_files, encrypted_data):
        with open(file, 'w') as f:
            f.write(encrypted)
    
    # Mark the project as encrypted
    project_config.set('encrypted', True)
    project_config.set('encryption-key-path', str(encryption_key))
    logger.info(f"Local files encrypted successfully with key {encryption_key}")






# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from pathlib import Path
from lean.components.util.logger import Logger
from base64 import b64decode, b64encode
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from lean.models.encryption import ActionType
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.models.api import QCProject, QCFullFile
from lean.components.config.storage import Storage
from lean.components.api.api_client import APIClient
from lean.components.util.organization_manager import OrganizationManager

def calculate_md5(input_string: str):
    """Calculate the md5 hash of a string

    :param input_string: The string to hash
    :return: The md5 hash of the string
    """
    from hashlib import md5
    return md5(input_string.encode()).hexdigest()

def get_b64_encoded(key: str) -> bytes:
    """Encode a string to base64

    :param key: The string to encode
    :return: The base64 encoded string
    """
    return b64encode(key.encode('utf-8'))

def get_project_key(project_key_path: Path, organization_id: str) -> str:
    """Get the project key from the project key file

    :param project_key_path: The path to the project key file
    :return: The project key
    """
    with open(project_key_path, 'r', encoding='utf-8', newline='') as f:
        content = f.read()
        key_for_aes = _get_fixed_length_key_from_user_full_length_key(content, organization_id.encode('utf-8'))
        return key_for_aes

def get_project_key_hash(project_key_path: Path):
    """Get the MD5 hash from the project key file

    :param project_key_path: The path to the project key file
    :return: The project iv
    """
    with open(project_key_path, 'r', encoding='utf-8', newline='') as f:
        content = f.read()
        return calculate_md5(content)

def get_project_iv(project_key_path: Path):
    """Get the project iv from the project key file

    :param project_key_path: The path to the project key file
    :return: The project iv
    """
    key_id = get_project_key_hash(project_key_path)
    return key_id[:16]

def are_encryption_keys_equal(key1: Path, key2: Path) -> bool:
    """Check if two encryption keys are equal

    :param key1: The first key to compare
    :param key2: The second key to compare
    :return: True if the keys are equal, False otherwise
    """
    if key1 is None and key2 is None:
        return True
    if key1 is None or key2 is None:
        return False
    return get_project_key_hash(key1) == get_project_key_hash(key2)

def get_decrypted_file_content_for_local_project(project: Path, source_files: List[Path], encryption_key: Path, project_config_manager: ProjectConfigManager, organization_id: str) -> List[str]:
    project_config = project_config_manager.get_project_config(project)

    # Check if the project is already encrypted
    areProjectFilesAlreadyEncrypted = project_config.get('encrypted', False)

    # Check if there is mismatch of keys
    _validate_key_state_for_local_project(project_config, encryption_key)

    project_key = get_project_key(encryption_key, organization_id)
    project_iv = get_project_iv(encryption_key)
    encoded_project_key = project_key.encode('utf-8')
    encoded_project_iv = project_iv.encode('utf-8')
    decrypted_data = []
    for file in source_files:
        try:
            # lets read and decrypt the file
            with open(file, 'r') as f:
                encrypted = f.read()
                if not areProjectFilesAlreadyEncrypted:
                    decrypted = encrypted
                else:
                    decrypted = _decrypt_file_content(encoded_project_key, encoded_project_iv, encrypted)
                decrypted_data.append(decrypted)
        except Exception as e:
            raise RuntimeError(f"Failed to decrypt file {file} with error {e}")
    return decrypted_data

def get_encrypted_file_content_for_local_project(project: Path, source_files: List[Path], encryption_key: Path, project_config_manager: ProjectConfigManager, organization_id: str) -> List[str]:
    project_config = project_config_manager.get_project_config(project)

    # Check if the project is already encrypted
    areProjectFilesAlreadyEncrypted = project_config.get('encrypted', False)

    # Check if there is mismatch of keys
    _validate_key_state_for_local_project(project_config, encryption_key)

    project_key = get_project_key(encryption_key, organization_id)
    project_iv = get_project_iv(encryption_key)
    encoded_project_key = project_key.encode('utf-8')
    encoded_project_iv = project_iv.encode('utf-8')
    encrypted_data: List[str] = []
    for file in source_files:
        try:
            # lets read and decrypt the file
            with open(file, 'r') as f:
                plain_text = f.read()
                if areProjectFilesAlreadyEncrypted:
                    encrypted = plain_text
                else:
                    encrypted = _encrypt_file_content(encoded_project_key, encoded_project_iv, plain_text.encode('utf-8'))
                encrypted_data.append(encrypted)
        except Exception as e:
            raise RuntimeError(f"Failed to encrypt file {file} with error {e}")
    return encrypted_data

def get_and_validate_user_input_encryption_key(user_input_key: Path, project_config_encryption_key: Path) -> str:
    if project_config_encryption_key is not None and Path(project_config_encryption_key).exists():
        project_config_encryption_key = Path(project_config_encryption_key)
    if project_config_encryption_key is None and user_input_key is None:
        raise RuntimeError("No encryption key was provided, please provide one using --key")
    elif project_config_encryption_key is None:
        project_config_encryption_key = user_input_key
    elif user_input_key is not None and project_config_encryption_key != user_input_key:
        raise RuntimeError(f"Provided encryption key ({user_input_key}) does not match the encryption key in the project ({project_config_encryption_key})")
    return project_config_encryption_key

def validate_user_inputs_for_cloud_push_pull_commands(encrypt: bool, decrypt: bool, key: Path):
    if encrypt and decrypt:
        raise RuntimeError(f"Cannot encrypt and decrypt at the same time.")
    if key is None and (encrypt or decrypt):
        raise RuntimeError(f"Encryption key is required when encrypting or decrypting.")
    if key is not None and not encrypt and not decrypt:
            raise RuntimeError(f"Encryption key can only be specified when encrypting or decrypting.")

def validate_encryption_key_registered_with_cloud(user_key: Path, organization_manager: OrganizationManager, api_client: APIClient):
    # lets check if the given key is registered with the cloud
    organization_id = organization_manager.try_get_working_organization_id()
    available_encryption_keys = api_client.encryption_keys.list(organization_id)['keys']
    encryption_key_id = get_project_key_hash(user_key)
    if (not any(found_key for found_key in available_encryption_keys if found_key['hash'] == encryption_key_id)):
        raise RuntimeError(f"Given encryption key is not registered with the cloud.")

def validate_key_and_encryption_state_for_cloud_project(project: QCProject, local_project_encryption_state: bool, encryption_key: Path, local_encryption_key: Path, logger:Logger) -> None:
    if not encryption_key and project.encryptionKey and local_encryption_key and local_encryption_key.exists() and get_project_key_hash(local_encryption_key) != project.encryptionKey.id:
        raise RuntimeError(f"Encryption Key mismatch. Local Project Key: {local_encryption_key}. Cloud Project Key: {project.encryptionKey.name}. Please provide correct encryption key for project '{project.name}' to proceed.")
    if not encryption_key and bool(project.encrypted) != bool(local_project_encryption_state):
        logger.debug(f"Force Overwrite: Project encryption state mismatch. Local Project Encrypted: {bool(local_project_encryption_state)}. Cloud Project Encrypted: {bool(project.encrypted)}.")
        return
    if encryption_key and project.encryptionKey and get_project_key_hash(encryption_key) != project.encryptionKey.id:
        raise RuntimeError(f"Encryption Key mismatch. Local Project Key hash: {get_project_key_hash(encryption_key)}. Cloud Project Key Hash: {project.encryptionKey.id}. Please provide correct encryption key for project '{project.name}' to proceed.")

def get_appropriate_files_from_cloud_project(project: QCProject, cloud_files: List[QCFullFile], encryption_key: Path, organization_id: str, encryption_action: ActionType) -> List[QCFullFile]:
    if encryption_action == ActionType.DECRYPT:
        return _get_decrypted_content_from_cloud_project(project, cloud_files, encryption_key, organization_id)
    return _get_encrypted_content_from_cloud_project(project, cloud_files, encryption_key, organization_id)

def get_appropriate_files_from_local_project(project: Path, paths: List[Path], encryption_key: Path, project_config_manager: ProjectConfigManager, organization_id: str, encryption_action: ActionType) -> List[str]:
    if encryption_action == ActionType.ENCRYPT:
        return get_encrypted_file_content_for_local_project(project, paths, encryption_key, project_config_manager, organization_id)
    return get_decrypted_file_content_for_local_project(project, paths, encryption_key,project_config_manager, organization_id)


def _validate_key_state_for_local_project(project_config: Storage, encryption_key: Path):
    local_encryption_key_path = project_config.get('encryption-key-path', None)
    if local_encryption_key_path and encryption_key and Path(local_encryption_key_path) != encryption_key:
        raise RuntimeError(f"Registered encryption key {local_encryption_key_path} is different from the one provided {encryption_key}")

def _decrypt_file_content(key: bytes, init_vector: bytes, b64_encoded_content: str) -> str:
    # remove new line characters that we added during encryption
    b64_encoded_content = b64_encoded_content.replace('\n', '')
    b64_encoded_content = b64_encoded_content.strip()
    content = b64decode(b64_encoded_content.encode('utf-8'))
    # Setup module-specific classes
    cipher = Cipher(algorithms.AES(key), modes.CBC(init_vector))
    decryptor = cipher.decryptor()

    plaintext = decryptor.update(content) + decryptor.finalize()
    # Unpad the decrypted data
    unpadder = padding.PKCS7(128).unpadder()
    unpadded_data = unpadder.update(plaintext) + unpadder.finalize()
    return unpadded_data.decode('utf-8').replace("\r\n", "\n")

def _encrypt_file_content(key: bytes, init_vector: bytes, content: bytes) -> str:
    plain_text = _pad(content, 16)
    # Setup module-specific classes
    cipher = Cipher(algorithms.AES(key), modes.CBC(init_vector))
    encryptor = cipher.encryptor()

    # Encrypt and decrypt data
    cipher_text = encryptor.update(plain_text) + encryptor.finalize()
    encrypted_content = b64encode(cipher_text).decode('utf-8')
    # let's make it user friendly and add new lines, same as local platform
    regex = r".{1,80}"
    from re import findall
    chunks = findall(regex, encrypted_content)
    return '\n'.join(chunks)

def _get_decrypted_content_from_cloud_project(project: QCProject, cloud_files: List[QCFullFile], encryption_key: Path, organization_id: str) -> List[QCFullFile]:
    # Check if the project is already encrypted
    if not project.encrypted or project.encryptionKey.id != get_project_key_hash(encryption_key):
        return cloud_files

    project_key = get_project_key(encryption_key, organization_id)
    project_iv = get_project_iv(encryption_key)
    encoded_project_key = project_key.encode('utf-8')
    encoded_project_iv = project_iv.encode('utf-8')
    for cloud_file in cloud_files:
        try:
            decrypted = _decrypt_file_content(encoded_project_key, encoded_project_iv, cloud_file.content)
            cloud_file.content = decrypted
        except Exception as e:
            raise RuntimeError(f"Failed to decrypt file {cloud_file} with error {e}")
    return cloud_files

def _get_encrypted_content_from_cloud_project(project: QCProject, cloud_files: List[QCFullFile], encryption_key: Path, organization_id: str) -> List[QCFullFile]:
    # Check if the project is already encrypted
    if project.encrypted:
        if encryption_key is not None and project.encryptionKey and project.encryptionKey.id != get_project_key_hash(encryption_key):
            raise RuntimeError(f"Registered encryption key {project.encryptionKey.id} is different from the one provided {encryption_key}")
        return cloud_files

    project_key = get_project_key(encryption_key, organization_id)
    project_iv = get_project_iv(encryption_key)
    encoded_project_key = project_key.encode('utf-8')
    encoded_project_iv = project_iv.encode('utf-8')
    for cloud_file in cloud_files:
        try:
            encrypted = _encrypt_file_content(encoded_project_key, encoded_project_iv, cloud_file.content.encode('utf-8'))
            cloud_file.content = encrypted
        except Exception as e:
            raise RuntimeError(f"Failed to decrypt file {cloud_file} with error {e}")
    return cloud_files

def _pad(data, block_size):
    """Padding function for AES encryption

    :param data: data to pad
    :param block_size: required block size
    :return: padded data
    """
    padding_length = block_size - (len(data) % block_size)
    return data + bytes([padding_length] * padding_length)

def _get_fixed_length_key_from_user_full_length_key(password: str, salt: bytes):
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=salt,
        iterations=100000,
        backend=default_backend()
    )
    return kdf.derive(password.encode()).hex()





# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

class TempManager:
    """The TempManager class provides access to temporary directories."""

    def __init__(self) -> None:
        """Creates a new TempManager instance."""
        self._temporary_directories = []
        self.delete_temporary_directories_when_done = True

    def create_temporary_directory(self) -> Path:
        """Returns the path to an empty temporary directory.

        :return: a path to an empty temporary directory
        """
        from tempfile import mkdtemp
        path = Path(mkdtemp(prefix="lean-cli-"))
        self._temporary_directories.append(path)
        return path

    def delete_temporary_directories(self) -> None:
        """Deletes temporary directories that were created while the CLI ran.

        Only the files that the user can delete are deleted, any permission errors are ignored.
        """
        from shutil import rmtree
        for path in self._temporary_directories:
            rmtree(path, ignore_errors=True)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any


def safe_save(data: str, path: Path, _retry: int = 0):
    from uuid import uuid4
    from shutil import move
    from os import unlink, path as os_path
    from time import time, sleep

    # make sure the parent folder exists
    path.parent.mkdir(parents=True, exist_ok=True)

    lock_file = Path(str(path) + '.lock').resolve()
    try:
        with open(lock_file, 'x') as _:
            # we create a tmp file we will use to write to. This was we avoid concurrency issues corrupting the file
            tmp_file = Path(str(path) + '.' + str(uuid4())).resolve()
            with open(tmp_file, "w+", encoding="utf-8") as targetFile:
                targetFile.write(data)

            # atomic move when on the same filesystem which is this case, they are side by side
            move(tmp_file, path)
        unlink(lock_file)
    except FileExistsError:
        try:
            # lock is taken, delete if old, else wait and retry
            if time() - os_path.getmtime(lock_file) >= 2:
                unlink(lock_file)
            else:
                sleep(0.250)
        except:
            # some else might have deleted it
            pass
        if _retry < 10:
            safe_save(data, path, ++_retry)


class Storage:
    """A Storage instance manages the data in a single JSON file."""

    def __init__(self, file: str) -> None:
        """Creates a new Storage instance.

        :param file: the path to the file this Storage instance should manage
        """
        from json import loads

        self.file = Path(file)

        if self.file.exists():
            try:
                content = self.file.read_text(encoding="utf-8")
                if content:
                    self._data = loads(content)
                else:
                    self._data = {}
            except:
                # Could happen rarely due to concurrency or unexpected failures, so if it does let's recover
                self.file.unlink()
                self._data = {}
        else:
            self._data = {}

    def is_empty(self) -> bool:
        """Determines if this storage file is empty

        :return: True if this storage file is empty
        """
        return len(self._data) == 0

    def get(self, key: str, default: Any = None) -> Any:
        """Returns the value assigned to the given key.

        Returns a default value when nothing is assigned to the given key.

        :param key: the key to retrieve the value of
        :param default: the default value to return when key is not set
        :return: the value assigned to the key or default if the key is not set
        """
        if key in self._data:
            return self._data[key]
        else:
            return default

    def set(self, key: str, value: Any) -> None:
        """Assigns a value to a key.

        The value is stored as json, so must be serializable using json.dump().

        :param key: the key to assign the value to
        :param value: the json-serializable value to assign to the given key
        """
        self._data[key] = value
        self._save()

    def delete(self, key: str) -> None:
        """Deletes a key.

        :param key: the key to delete
        """
        self._data.pop(key, None)
        self._save()

    def has(self, key: str) -> bool:
        """Returns whether the Storage instance has a value assigned to the given key.

        :param key: the key to check the existence of
        :return: True if a value is assigned to the given key, False if not
        """
        return key in self._data

    def clear(self) -> None:
        """Clears the Storage instance and deletes the underlying file."""
        self._data.clear()
        self._save()

    def _save(self) -> None:
        from json import dumps

        """Saves the data to the underlying file, deleting the file if there is no data."""
        if len(self._data) > 0:
            safe_save(data=dumps(self._data, indent=4) + "\n", path=self.file.resolve())
        else:
            if self.file.exists():
                self.file.unlink()



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from click import group

from lean.commands.cloud.backtest import backtest
from lean.commands.cloud.live.live import live
from lean.commands.cloud.optimize import optimize
from lean.commands.cloud.pull import pull
from lean.commands.cloud.push import push
from lean.commands.cloud.status import status
from lean.commands.cloud.object_store import object_store

@group()
def cloud() -> None:
    """Interact with the QuantConnect cloud."""
    # This method is intentionally empty
    # It is used as the command group for all `lean cloud <command>` commands
    pass


cloud.add_command(pull)
cloud.add_command(push)
cloud.add_command(backtest)
cloud.add_command(optimize)
cloud.add_command(live)
cloud.add_command(status)
cloud.add_command(object_store)


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lean.components.config.storage import Storage
from lean.components.docker.docker_manager import DockerManager
from lean.components.util.http_client import HTTPClient
from lean.components.util.logger import Logger
from lean.constants import (UPDATE_CHECK_INTERVAL_ANNOUNCEMENTS, UPDATE_CHECK_INTERVAL_CLI,
                            UPDATE_CHECK_INTERVAL_DOCKER_IMAGE)
from lean.models.docker import DockerImage


class UpdateManager:
    """The UpdateManager class contains methods to check for and warn the user about available updates."""

    def __init__(self,
                 logger: Logger,
                 http_client: HTTPClient,
                 cache_storage: Storage,
                 docker_manager: DockerManager) -> None:
        """Creates a new UpdateManager instance.

        :param logger: the logger to use when warning the user when something is outdated
        :param http_client: the HTTPClient instance to use for HTTP requests
        :param cache_storage: the Storage instance to use for getting/setting the last time a certain update check was performed
        :param docker_manager: the DockerManager instance to use to check for Docker updates
        """
        self._logger = logger
        self._http_client = http_client
        self._cache_storage = cache_storage
        self._docker_manager = docker_manager

    def warn_if_cli_outdated(self, force: bool = False) -> None:
        """Warns the user if the CLI is outdated.

        An update check is performed once every UPDATE_CHECK_INTERVAL_CLI hours, unless force is set to True.

        :param force: whether the update check interval should be bypassed (defaults to False)
        """
        from lean import __version__
        current_version = __version__

        # A development version is never considered outdated
        if "dev" in current_version:
            return

        if not force and not self._should_check_for_updates("cli", UPDATE_CHECK_INTERVAL_CLI):
            return

        from requests import exceptions
        try:
            response = self._http_client.get("https://pypi.org/pypi/lean/json", raise_for_status=False)
        except exceptions.ConnectionError:
            # The user may be offline, do nothing
            return

        if not response.ok:
            return

        latest_version = response.json()["info"]["version"]
        from distutils.version import StrictVersion

        if StrictVersion(latest_version) > StrictVersion(current_version):
            self._logger.warn(f"A new release of the Lean CLI is available ({current_version} -> {latest_version})")
            self._logger.warn("Run `pip install --upgrade lean` to update to the latest version")

    def pull_docker_image_if_necessary(self, image: DockerImage, force: bool, no_update = False) -> None:
        """Pulls a Docker image if necessary.

        Docker images are pulled when they are not installed yet,
        and once every UPDATE_CHECK_INTERVAL_DOCKER_IMAGE hours (the interval is per image).

        :param image: the image to pull
        :param force: skip the interval check to force a pull
        :param no_update: do not pull the image
        """

        if no_update:
            if not force:
                return
            self._logger.warn("Both --no-update and --update were specified, ignoring --no-update")

        from docker.errors import APIError
        if not force and self._docker_manager.image_installed(image):
            if not self._should_check_for_updates(str(image), UPDATE_CHECK_INTERVAL_DOCKER_IMAGE):
                return

            local_digest = self._docker_manager.get_local_digest(image)

            # If the local digest does not exist the image was built locally so there is nothing to pull
            if local_digest is None:
                return

            try:
                remote_digest = self._docker_manager.get_remote_digest(image)
            except APIError:
                # The user may be offline, do nothing
                return

            # Do nothing if we're already using the latest image
            if remote_digest == local_digest:
                return

        self._docker_manager.pull_image(image)

    def show_announcements(self) -> None:
        """Shows the announcements if they have been updated.

        We check for new announcements once every UPDATE_CHECK_INTERVAL_ANNOUNCEMENTS hours.
        """
        from requests import exceptions
        from hashlib import md5
        from rich import box
        from rich.panel import Panel
        from rich.table import Table

        if not self._should_check_for_updates("announcements", UPDATE_CHECK_INTERVAL_ANNOUNCEMENTS):
            return

        try:
            response = self._http_client.get(
                "https://raw.githubusercontent.com/QuantConnect/lean-cli/master/announcements.json",
                raise_for_status=False
            )
        except exceptions.ConnectionError:
            # The user may be offline, do nothing
            return

        if not response.ok:
            return

        hash_cache_key = "last-announcements-hash"

        remote_hash = md5(response.content).hexdigest()
        local_hash = self._cache_storage.get(hash_cache_key, None)

        if local_hash == remote_hash:
            return

        self._cache_storage.set(hash_cache_key, remote_hash)

        announcements = response.json()["announcements"]
        if len(announcements) == 0:
            return

        table = Table.grid(padding=(0, 1))
        table.add_column(overflow="fold")
        table.add_column(overflow="fold", ratio=1)

        for announcement in announcements:
            table.add_row(announcement["date"] + ":", announcement["message"])

        self._logger.info(Panel.fit(table, title="Announcements", box=box.SQUARE))

    def _should_check_for_updates(self, key: str, interval_hours: int) -> bool:
        """Returns whether an update check should be performed.

        When this method returns True, for the next <interval_hours> hours it returns False for <key>.

        :param key: the key which is used to identify the current update check
        :param interval_hours: the amount of hours between update checks for the given key
        :return: True if an update check should be performed, False if not
        """
        from datetime import datetime, timedelta, timezone

        storage_key = f"last-update-check-{key}"
        should_check = False

        if not self._cache_storage.has(storage_key):
            # Perform an update check if we haven't ran one yet
            should_check = True
        else:
            last_update_check = datetime.fromtimestamp(self._cache_storage.get(storage_key), tz=timezone.utc)
            time_since_last_update_check = datetime.now(tz=timezone.utc) - last_update_check

            if time_since_last_update_check >= timedelta(hours=interval_hours):
                # Perform an update check if the time since the last update check is greater than the given interval
                should_check = True

        if should_check:
            # Save the current timestamp so for the next <interval_hours> hours we can return False for <key>
            self._cache_storage.set(storage_key, datetime.now(tz=timezone.utc).timestamp())

        return should_check



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from lean.components import reserved_names, forbidden_characters
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.util.platform_manager import PlatformManager


class PathManager:
    """The PathManager class provides utilities for working with paths."""

    def __init__(self, lean_config_manager: LeanConfigManager, platform_manager: PlatformManager) -> None:
        """Creates a new PathManager instance.

        :param platform_manager: the PlatformManager used when checking which operating system is in use
        """
        self._lean_config_manager = lean_config_manager
        self._platform_manager = platform_manager

    def get_relative_path(self, destination: Path, source: Path = Path.cwd()) -> Path:
        """Returns a path relative to another one.

        :param destination: the path to point to
        :param source: the root where the relative path is relative to
        :return: the destination path relative to the source path, or destination path if it is not relative
        """
        try:
            return destination.relative_to(source)
        except ValueError:
            return destination

    def is_name_valid(self, name: str) -> bool:
        """Returns whether a name is valid on Windows operating system.

        :param name: the name to validate
        :return: True if the name is valid on Windows operating system, False if not
        """
        import re
        return re.match(r'^[-_a-zA-Z0-9/\s]*$', name) is not None

    def is_path_valid(self, path: Path) -> bool:
        """Returns whether the given path is a valid project path in the current operating system.

        This method should only be used to check paths relative to the current lean init folder.
        Passing an absolute path might result in false being returned since especial cases for root directories
        for each operating system (like devices in Windows) are not validated.

        :param path: the path to validate
        :return: True if the path is valid on the current operating system, False if not
        """
        try:
            # This call fails if the path contains invalid characters
            path.exists()
        except OSError:
            return False

        # On Windows path.exists() doesn't throw for paths like CON/file.txt
        # Trying to create them does raise errors, so we manually validate path components
        # We follow the rules of windows for every OS
        components = path.as_posix().split("/")
        for component in components:
            if component.startswith(" ") or component.endswith(" ") or component.endswith("."):
                return False

            for reserved_name in reserved_names:
                if component.upper() == reserved_name or component.upper().startswith(reserved_name + "."):
                    return False

            for forbidden_character in forbidden_characters:
                if forbidden_character in component:
                    return False
        return True

    def is_cli_path_valid(self, path: Path) -> bool:
        """Returns whether the given path is a valid project path in the current operating system.

        :param path: the path to validate
        :return: True if the path is valid on the current operating system, False if not
        """
        from lean.models.errors import MoreInfoError

        relative_path = path

        try:
            cli_root_dir = self._lean_config_manager.get_cli_root_directory()
            relative_path = path.relative_to(cli_root_dir)
        except (MoreInfoError, ValueError):
            from platform import system
            if system() == "Windows":
                # Skip the first component, which contains the drive name
                posix_path = path.as_posix()
                first_separator_index = posix_path.find('/')
                relative_path = Path(posix_path[first_separator_index:] if first_separator_index != -1 else path)

        return relative_path == Path(".") or self.is_path_valid(relative_path)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional, List, Tuple
from datetime import datetime, timedelta

from click import command, argument, option, Choice, IntRange

from lean.click import LeanCommand, PathParameter, ensure_options
from lean.components.docker.lean_runner import LeanRunner
from lean.constants import DEFAULT_ENGINE_IMAGE
from lean.container import container
from lean.models.api import QCParameter, QCBacktest
from lean.models.click_options import options_from_json, get_configs_for_options
from lean.models.cli import cli_data_downloaders, cli_addon_modules
from lean.models.errors import MoreInfoError
from lean.models.optimizer import OptimizationTarget
from lean.components.util.json_modules_handler import build_and_configure_modules, non_interactive_config_build_for_name


def _get_latest_backtest_runtime(algorithm_directory: Path) -> timedelta:
    from re import findall
    from dateutil.parser import isoparse

    missing_backtest_error = RuntimeError(
        "Please run at least one backtest for this project in order to run an optimization estimate");
    backtests_directory = algorithm_directory / "backtests"

    if not backtests_directory.exists():
        raise missing_backtest_error

    def is_backtest_output_directory(path: Path) -> bool:
        try:
            datetime.strptime(path.name, "%Y-%m-%d_%H-%M-%S")
        except ValueError:
            return False
        else:
            return path.is_dir()

    def get_filename_timestamp(path: Path) -> datetime:
        return datetime.strptime(path.name, "%Y-%m-%d_%H-%M-%S")

    backtests_directories = [f for f in backtests_directory.iterdir() if is_backtest_output_directory(f)]
    latest_backtest_directory = max(backtests_directories, key=get_filename_timestamp, default=None)

    if latest_backtest_directory is None:
        raise missing_backtest_error

    latest_backtest_log_file = latest_backtest_directory / "log.txt"

    if not latest_backtest_log_file.exists():
        raise missing_backtest_error

    latest_backtest_logs = latest_backtest_log_file.read_text(encoding="utf-8")
    timestamps = findall(r"(.+) TRACE:: .*\n", latest_backtest_logs)

    return isoparse(timestamps[-1]) - isoparse(timestamps[0])


@command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@option("--output",
              type=PathParameter(exists=False, file_okay=False, dir_okay=True),
              help="Directory to store results in (defaults to PROJECT/optimizations/TIMESTAMP)")
@option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run the optimization in a detached Docker container and return immediately")
@option("--optimizer-config",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help=f"The optimizer configuration file that should be used")
@option("--strategy",
              type=Choice(["Grid Search", "Euler Search"], case_sensitive=False),
              help="The optimization strategy to use")
@option("--target",
              type=str,
              help="The target statistic of the optimization")
@option("--target-direction",
              type=Choice(["min", "max"], case_sensitive=False),
              help="Whether the target must be minimized or maximized")
@option("--parameter",
              type=(str, float, float, float),
              multiple=True,
              help="The 'parameter min max step' pairs configuring the parameters to optimize")
@option("--constraint",
              type=str,
              multiple=True,
              help="The 'statistic operator value' pairs configuring the constraints of the optimization")
@option("--data-provider-historical",
              type=Choice([dp.get_name() for dp in cli_data_downloaders], case_sensitive=False),
              default="Local",
              help="Update the Lean configuration file to retrieve data from the given historical provider")
@option("--download-data",
              is_flag=True,
              default=False,
              help="Update the Lean configuration file to download data from the QuantConnect API, alias for --data-provider-historical QuantConnect")
@option("--data-purchase-limit",
              type=int,
              help="The maximum amount of QCC to spend on downloading data during the backtest when using QuantConnect as historical data provider")
@option("--release",
              is_flag=True,
              default=False,
              help="Compile C# projects in release configuration instead of debug")
@option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before running the optimizer")
@option("--estimate",
              is_flag=True,
              default=False,
              help="Estimate optimization runtime without running it")
@option("--max-concurrent-backtests",
              type=IntRange(min=1),
              help="Maximum number of concurrent backtests to run")
@option("--addon-module",
              type=str,
              multiple=True,
              hidden=True)
@option("--extra-config",
              type=(str, str),
              multiple=True,
              hidden=True)
@option("--extra-docker-config",
              type=str,
              default="{}",
              help="Extra docker configuration as a JSON string. "
                   "For more information https://docker-py.readthedocs.io/en/stable/containers.html")
@option("--no-update",
              is_flag=True,
              default=False,
              help="Use the local LEAN engine image instead of pulling the latest version")
@options_from_json(get_configs_for_options("backtest"))
def optimize(project: Path,
             output: Optional[Path],
             detach: bool,
             optimizer_config: Optional[Path],
             strategy: Optional[str],
             target: Optional[str],
             target_direction: Optional[str],
             parameter: List[Tuple[str, float, float, float]],
             constraint: List[str],
             data_provider_historical: Optional[str],
             download_data: bool,
             data_purchase_limit: Optional[int],
             release: bool,
             image: Optional[str],
             update: bool,
             estimate: bool,
             max_concurrent_backtests: Optional[int],
             addon_module: Optional[List[str]],
             extra_config: Optional[Tuple[str, str]],
             extra_docker_config: Optional[str],
             no_update: bool,
             **kwargs) -> None:
    """Optimize a project's parameters locally using Docker.

    \b
    If PROJECT is a directory, the algorithm in the main.py or Main.cs file inside it will be executed.
    If PROJECT is a file, the algorithm in the specified file will be executed.

    By default an interactive wizard is shown letting you configure the optimizer.
    If --optimizer-config or --strategy is given the command runs in non-interactive mode.
    In this mode the CLI does not prompt for input.

    \b
    The --optimizer-config option can be used to specify the configuration to run the optimizer with.
    When using the option it should point to a file like this (the algorithm-* properties should be omitted):
    https://github.com/QuantConnect/Lean/blob/master/Optimizer.Launcher/config.json

    If --strategy is given the optimizer configuration is read from the given options.
    In this case --strategy, --target, --target-direction and --parameter become required.

    \b
    In non-interactive mode the --parameter option can be provided multiple times to configure multiple parameters:
    - --parameter <name> <min value> <max value> <step size>
    - --parameter my-first-parameter 1 10 0.5 --parameter my-second-parameter 20 30 5

    \b
    In non-interactive mode the --constraint option can be provided multiple times to configure multiple constraints:
    - --constraint "<statistic> <operator> <value>"
    - --constraint "Sharpe Ratio >= 0.5" --constraint "Drawdown < 0.25"

    \b
    If --estimate is given, the optimization will not be executed.
    The runtime estimate for the optimization will be calculated and outputted.

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    from json import dumps
    from json5 import loads
    from docker.types import Mount
    from re import findall, search
    from os import cpu_count
    from math import floor

    should_detach = detach and not estimate
    environment_name = "backtesting"
    project_manager = container.project_manager
    algorithm_file = project_manager.find_algorithm_file(project)

    latest_backtest_runtime = timedelta(0)
    if estimate:
        latest_backtest_runtime = _get_latest_backtest_runtime(algorithm_file.parent)

    if output is None:
        output = algorithm_file.parent / "optimizations" / datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    optimizer_config_manager = container.optimizer_config_manager
    config = None

    if optimizer_config is not None and strategy is not None:
        raise RuntimeError("--optimizer-config and --strategy are mutually exclusive")

    if optimizer_config is not None:
        config = loads(optimizer_config.read_text(encoding="utf-8"))

        # Remove keys which are configured in the Lean config
        for key in ["algorithm-type-name", "algorithm-language", "algorithm-location"]:
            config.pop(key, None)
    elif strategy is not None:
        ensure_options(["strategy", "target", "target_direction", "parameter"])

        optimization_strategy = f"QuantConnect.Optimizer.Strategies.{strategy.replace(' ', '')}OptimizationStrategy"
        optimization_target = OptimizationTarget(target=optimizer_config_manager.parse_target(target),
                                                 extremum=target_direction)
        optimization_parameters = optimizer_config_manager.parse_parameters(parameter)
        optimization_constraints = optimizer_config_manager.parse_constraints(constraint)
    else:
        project_config_manager = container.project_config_manager
        project_config = project_config_manager.get_project_config(algorithm_file.parent)
        project_parameters = [QCParameter(key=k, value=v) for k, v in project_config.get("parameters", {}).items()]

        if len(project_parameters) == 0:
            raise MoreInfoError("The given project has no parameters to optimize",
                                "https://www.lean.io/docs/v2/lean-cli/optimization/parameters")

        optimization_strategy = optimizer_config_manager.configure_strategy(cloud=False)
        optimization_target = optimizer_config_manager.configure_target()
        optimization_parameters = optimizer_config_manager.configure_parameters(project_parameters, cloud=False)
        optimization_constraints = optimizer_config_manager.configure_constraints()

    if config is None:
        # noinspection PyUnboundLocalVariable
        config = {
            "optimization-strategy": optimization_strategy,
            "optimization-strategy-settings": {
                "$type": "QuantConnect.Optimizer.Strategies.StepBaseOptimizationStrategySettings, QuantConnect.Optimizer",
                "default-segment-amount": 10
            },
            "optimization-criterion": {
                "target": optimization_target.target,
                "extremum": optimization_target.extremum.value
            },
            "parameters": [parameter.dict() for parameter in optimization_parameters],
            "constraints": [constraint.dict(by_alias=True) for constraint in optimization_constraints]
        }

    if max_concurrent_backtests is not None:
        config["maximum-concurrent-backtests"] = max_concurrent_backtests
    elif "maximum-concurrent-backtests" in config:
        max_concurrent_backtests = config["maximum-concurrent-backtests"]
    else:
        max_concurrent_backtests = max(1, floor(cpu_count() / 2))

    config["optimizer-close-automatically"] = True
    config["results-destination-folder"] = "/Results"

    config_path = output / "optimizer-config.json"
    config_path.parent.mkdir(parents=True, exist_ok=True)
    with config_path.open("w+", encoding="utf-8") as file:
        file.write(dumps(config, indent=4) + "\n")

    project_config_manager = container.project_config_manager
    cli_config_manager = container.cli_config_manager

    project_config = project_config_manager.get_project_config(algorithm_file.parent)
    engine_image = cli_config_manager.get_engine_image(image or project_config.get("engine-image", None))

    logger = container.logger

    if str(engine_image) != DEFAULT_ENGINE_IMAGE:
        logger.warn(f'A custom engine image: "{engine_image}" is being used!')

    lean_config_manager = container.lean_config_manager
    lean_config = lean_config_manager.get_complete_lean_config(environment_name, algorithm_file, None)

    organization_id = container.organization_manager.try_get_working_organization_id()

    if download_data:
        data_provider_historical = "QuantConnect"

    paths_to_mount = None

    if data_provider_historical is not None:
        data_provider = non_interactive_config_build_for_name(lean_config, data_provider_historical,
                                                              cli_data_downloaders, kwargs, logger, environment_name)
        data_provider.ensure_module_installed(organization_id)
        container.lean_config_manager.set_properties(data_provider.get_settings())
        paths_to_mount = data_provider.get_paths_to_mount()

    lean_config_manager.configure_data_purchase_limit(lean_config, data_purchase_limit)

    if not output.exists():
        output.mkdir(parents=True)

    # This maybe overwritten by the addon module later if given.
    lean_config["messaging-handler"] = "QuantConnect.Messaging.Messaging"

    lean_runner = container.lean_runner

    # Set extra config
    for key, value in extra_config:
        if "environments" in lean_config and environment_name in lean_config["environments"] \
                and key in lean_config["environments"][environment_name]:
            lean_config["environments"][environment_name][key] = value
        else:
            lean_config[key] = value

    output_config_manager = container.output_config_manager
    lean_config["algorithm-id"] = str(output_config_manager.get_optimization_id(output))

    # Configure addon modules
    build_and_configure_modules(addon_module, cli_addon_modules, organization_id, lean_config,
                                kwargs, logger, environment_name)

    container.update_manager.pull_docker_image_if_necessary(engine_image, update, no_update)

    run_options = lean_runner.get_basic_docker_config(lean_config, algorithm_file, output, None, release, should_detach,
                                                      engine_image, paths_to_mount)

    run_options["working_dir"] = "/Lean/Optimizer.Launcher/bin/Debug"
    run_options["commands"].append(f"dotnet QuantConnect.Optimizer.Launcher.dll{' --estimate' if estimate else ''}")
    run_options["mounts"].append(
        Mount(target="/Lean/Optimizer.Launcher/bin/Debug/config.json",
              source=str(config_path),
              type="bind",
              read_only=True)
    )

    # Add known additional run options from the extra docker config
    LeanRunner.parse_extra_docker_config(run_options, loads(extra_docker_config))

    project_manager.copy_code(algorithm_file.parent, output / "code")

    success = container.docker_manager.run_image(engine_image, **run_options)

    cli_root_dir = container.lean_config_manager.get_cli_root_directory()
    relative_project_dir = project.relative_to(cli_root_dir)
    relative_output_dir = output.relative_to(cli_root_dir)

    if should_detach:
        temp_manager = container.temp_manager
        temp_manager.delete_temporary_directories_when_done = False

        logger.info(
            f"Successfully started optimization for '{relative_project_dir}' in the '{run_options['name']}' container")
        logger.info(f"The output will be stored in '{relative_output_dir}'")
        logger.info("You can use Docker's own commands to manage the detached container")
    elif success:
        optimizer_logs = (output / "log.txt").read_text(encoding="utf-8")

        if estimate:
            match = search(r"Optimization estimate: (\d+)", optimizer_logs)

            if match is None:
                raise RuntimeError(f"Something went wrong while running the optimization estimate, "
                                   f"the output is stored in '{relative_output_dir}'")

            backtestsCount = int(match[1])
            logger.info(f"Optimization estimate: \n"
                        f"  Total backtests: {backtestsCount}\n"
                        f"  Estimated runtime: {backtestsCount * latest_backtest_runtime / max_concurrent_backtests}")
        else:
            groups = findall(r"ParameterSet: \(([^)]+)\) backtestId '([^']+)'", optimizer_logs)

            if len(groups) > 0:
                optimal_parameters, optimal_id = groups[0]

                optimal_results = loads((output / optimal_id / f"{optimal_id}.json").read_text(encoding="utf-8"))
                optimal_backtest = QCBacktest(backtestId=optimal_id,
                                              projectId=1,
                                              status="",
                                              name=optimal_id,
                                              created=datetime.now(),
                                              completed=True,
                                              progress=1.0,
                                              runtimeStatistics=optimal_results["RuntimeStatistics"],
                                              statistics=optimal_results["Statistics"])

                logger.info(f"Optimal parameters: {optimal_parameters.replace(':', ': ').replace(',', ', ')}")
                logger.info(f"Optimal backtest results:")
                logger.info(optimal_backtest.get_statistics_table())

            logger.info(
                f"Successfully optimized '{relative_project_dir}' and stored the output in '{relative_output_dir}'")
    else:
        raise RuntimeError(
            f"Something went wrong while running the optimization, the output is stored in '{relative_output_dir}'")
